from .constraints import RecordsSchemaFieldConstraints
from typing import Optional, cast, TYPE_CHECKING
if TYPE_CHECKING:
    import sqlalchemy
    from records_mover.db import DBDriver  # noqa
    from .constraints import FieldIntegerConstraintsDict  # noqa


class RecordsSchemaFieldIntegerConstraints(RecordsSchemaFieldConstraints):
    def __init__(self,
                 required: bool,
                 unique: Optional[bool],
                 min_: Optional[int],
                 max_: Optional[int]):
        super().__init__(required=required, unique=unique)
        self.min_ = min_
        self.max_ = max_

    @staticmethod
    def from_sqlalchemy_type(required: bool,
                             unique: Optional[bool],
                             type_: 'sqlalchemy.types.TypeEngine',
                             driver: 'DBDriver') -> 'RecordsSchemaFieldIntegerConstraints':
        import sqlalchemy

        if not isinstance(type_, sqlalchemy.types.Integer):
            raise TypeError(f"Unexpected column type: {type_}")
        limits = driver.integer_limits(type_)
        min_: Optional[int] = None
        max_: Optional[int] = None
        if limits:
            min_, max_ = limits

        return RecordsSchemaFieldIntegerConstraints(required=required,
                                                    unique=unique,
                                                    min_=min_,
                                                    max_=max_)

    def to_data(self) -> 'FieldIntegerConstraintsDict':
        raw_out = super().to_data()
        out = cast('FieldIntegerConstraintsDict', raw_out)
        if self.min_ is not None:
            out['min'] = str(self.min_)
        if self.max_ is not None:
            out['max'] = str(self.max_)
        return out



from typing import Dict, Union, NoReturn, TYPE_CHECKING
from .mode import CopyOptionsMode
if TYPE_CHECKING:
    from typing_extensions import Literal
    # https://www.postgresql.org/docs/9.5/runtime-config-client.html#GUC-DATESTYLE
    #
    # DateStyle (string)
    #
    #  Sets the display format for date and time values, as well as
    #  the rules for interpreting ambiguous date input values. For
    #  historical reasons, this variable contains two independent
    #  components: the output format specification (ISO, Postgres,
    #  SQL, or German) and the input/output specification for
    #  year/month/day ordering (DMY, MDY, or YMD). These can be set
    #  separately or together. The keywords Euro and European are
    #  synonyms for DMY; the keywords US, NonEuro, and NonEuropean are
    #  synonyms for MDY. See Section 8.5 for more information. The
    #  built-in default is ISO, MDY, but initdb will initialize the
    #  configuration file with a setting that corresponds to the
    #  behavior of the chosen lc_time locale.
    DateOrderStyle = Union[Literal["DMY"], Literal["MDY"]]

    # https://www.postgresql.org/docs/9.5/datatype-datetime.html#DATATYPE-DATETIME-OUTPUT
    #
    # The output format of the date/time types can be set to one of
    # the four styles ISO 8601, SQL (Ingres), traditional POSTGRES
    # (Unix date format), or German. The default is the ISO
    # format. (The SQL standard requires the use of the ISO 8601
    # format. The name of the "SQL" output format is a historical
    # accident.) Table 8-14 shows examples of each output style. The
    # output of the date and time types is generally only the date or
    # time part in accordance with the given examples. However, the
    # POSTGRES style outputs date-only values in ISO format.

    # Style Specification      Description                Example
    # ISO                      ISO 8601, SQL standard     1997-12-17 07:37:16-08
    # SQL                      traditional style          12/17/1997 07:37:16.00 PST
    # Postgres                 original style             Wed Dec 17 07:37:16 1997 PST
    # German                   regional style             17.12.1997 07:37:16.00 PST

    # Note: ISO 8601 specifies the use of uppercase letter T to
    # separate the date and time. PostgreSQL accepts that format on
    # input, but on output it uses a space rather than T, as shown
    # above. This is for readability and for consistency with RFC 3339
    # as well as some other database systems.

    # In the SQL and POSTGRES styles, day appears before month if DMY
    # field ordering has been specified, otherwise month appears
    # before day. (See Section 8.5.1 for how this setting also affects
    # interpretation of input values.) Table 8-15 shows examples.

    # datestyle Setting          Input Ordering     Example Output
    # SQL, DMY                   day/month/year     17/12/1997 15:37:16.00 CET
    # SQL, MDY                   month/day/year     12/17/1997 07:37:16.00 PST
    # Postgres, DMY              day/month/year     Wed 17 Dec 07:37:16 1997 PST

    # The date/time style can be selected by the user using the SET
    # datestyle command, the DateStyle parameter in the
    # postgresql.conf configuration file, or the PGDATESTYLE
    # environment variable on the server or client.

    # The formatting function to_char (see Section 9.8) is also
    # available as a more flexible way to format date/time output.
    DateOutputStyle = Literal["ISO", "SQL", "Postgres", "German"]

    CopyOptionsModeType = Union[Literal[CopyOptionsMode.LOADING],
                                Literal[CopyOptionsMode.UNLOADING]]
else:
    DateOrderStyle = str
    DateOutputStyle = str
    CopyOptionsModeType = str


PostgresCopyOptions = Dict[str, object]


# mypy way of validating we're covering all cases of an enum
#
# https://github.com/python/mypy/issues/6366#issuecomment-560369716
def _assert_never(x: NoReturn) -> NoReturn:
    assert False, "Unhandled type: {}".format(type(x).__name__)



from ..driver import DBDriver
import sqlalchemy
from sqlalchemy.schema import Table
from records_mover.records import RecordsSchema
from records_mover.records.records_format import BaseRecordsFormat, AvroRecordsFormat
import logging
from ...utils.limits import (INT16_MIN, INT16_MAX,
                             INT32_MIN, INT32_MAX,
                             INT64_MIN, INT64_MAX,
                             FLOAT32_SIGNIFICAND_BITS,
                             FLOAT64_SIGNIFICAND_BITS,
                             num_digits)
from .sql import schema_sql_from_admin_views
import timeout_decorator
from typing import Optional, Union, Dict, List, Tuple
from ...url.base import BaseDirectoryUrl
from records_mover.db.quoting import quote_group_name, quote_schema_and_table
from .unloader import RedshiftUnloader
from ..unloader import Unloader
from .loader import RedshiftLoader
from ..loader import LoaderFromRecordsDirectory


logger = logging.getLogger(__name__)


class RedshiftDBDriver(DBDriver):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
                 s3_temp_base_loc: Optional[BaseDirectoryUrl] = None,
                 **kwargs) -> None:
        super().__init__(db)
        self.s3_temp_base_loc = s3_temp_base_loc
        self._redshift_loader =\
            RedshiftLoader(db=db,
                           meta=self.meta,
                           s3_temp_base_loc=s3_temp_base_loc)
        self._redshift_unloader =\
            RedshiftUnloader(db=db,
                             table=self.table,
                             s3_temp_base_loc=s3_temp_base_loc)

    def schema_sql(self, schema: str, table: str) -> str:
        out = schema_sql_from_admin_views(schema, table, self.db)
        if out is None:
            return super().schema_sql(schema=schema, table=table)
        else:
            return out

    # if this timeout goes off (at least for Redshift), it's probably
    # because memory is filling because sqlalchemy's cache of all
    # tables and columns filled up memory in the job.
    @timeout_decorator.timeout(80)
    def table(self, schema: str, table: str) -> Table:
        with self.db.engine.connect() as conn:
            with conn.begin():
                # The code in the Redshift SQLAlchemy driver relies on 'SET
                # LOCAL search_path' in order to do reflection and pull back
                # column information on tables.
                #
                # Unfortunately, 'SET LOCAL' only works when a transaction is
                # in process, and it looks like there's often not here, and an
                # exception will be thrown unless the table you're working
                # with is in your default schema.
                #
                # There was some hang-wringing when they added it originally:
                #  https://github.com/sqlalchemy-redshift/sqlalchemy-redshift/commit/
                #      94723ec6437c5e5197fcf785845499e81640b167
                return Table(table, self.meta, schema=schema, autoload=True, autoload_with=conn)

    def set_grant_permissions_for_groups(self, schema_name: str, table: str,
                                         groups: Dict[str, List[str]],
                                         db: Union[sqlalchemy.engine.Engine,
                                                   sqlalchemy.engine.Connection]) -> None:
        schema_and_table = quote_schema_and_table(self.db.engine, schema_name, table)
        for perm_type in groups:
            groups_list = groups[perm_type]
            for group in groups_list:
                group_name: str = quote_group_name(self.db.engine, group)
                if not perm_type.isalpha():
                    raise TypeError("Please make sure your permission types"
                                    " are an acceptable value.")
                perms_sql = f'GRANT {perm_type} ON TABLE {schema_and_table} TO GROUP {group_name}'
                db.execute(perms_sql)
        return None

    def supports_time_type(self) -> bool:
        return False

    def integer_limits(self,
                       type_: sqlalchemy.types.Integer) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.sql.sqltypes.SMALLINT):
            return (INT16_MIN, INT16_MAX)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.INTEGER):
            return (INT32_MIN, INT32_MAX)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.BIGINT):
            return (INT64_MIN, INT64_MAX)
        return super().integer_limits(type_)

    def fp_constraints(self,
                       type_: sqlalchemy.types.Float) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.dialects.postgresql.base.DOUBLE_PRECISION):
            return (64, FLOAT64_SIGNIFICAND_BITS)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.REAL):
            return (32, FLOAT32_SIGNIFICAND_BITS)
        return super().fp_constraints(type_)

    def type_for_integer(self,
                         min_value: Optional[int],
                         max_value: Optional[int]) -> sqlalchemy.types.TypeEngine:
        """Find correct integral column type to fit the given min and max integer values"""

        if min_value is not None and max_value is not None:
            if min_value >= INT16_MIN and max_value <= INT16_MAX:
                return sqlalchemy.sql.sqltypes.SMALLINT()
            if min_value >= INT32_MIN and max_value <= INT32_MAX:
                return sqlalchemy.sql.sqltypes.INTEGER()
            if min_value >= INT64_MIN and max_value <= INT64_MAX:
                return sqlalchemy.sql.sqltypes.BIGINT()
            else:
                num_digits_min = num_digits(min_value)
                num_digits_max = num_digits(max_value)
                digit_count = max(num_digits_min, num_digits_max)
                return self.type_for_fixed_point(precision=digit_count,
                                                 scale=0)
        return super().type_for_integer(min_value, max_value)

    def type_for_fixed_point(self,
                             precision: int,
                             scale: int) -> sqlalchemy.sql.sqltypes.Numeric:
        if precision > 38:
            logger.warning("Using floating point type, as Redshift can't store >38 "
                           f"digits of precision in a fixed point type (requested: {precision})")
            return sqlalchemy.dialects.postgresql.base.DOUBLE_PRECISION()
        return super().type_for_fixed_point(precision=precision,
                                            scale=scale)

    def type_for_floating_point(self,
                                fp_total_bits: int,
                                fp_significand_bits: int) -> sqlalchemy.sql.sqltypes.Numeric:
        if fp_significand_bits > FLOAT64_SIGNIFICAND_BITS:
            logger.warning(f"Falling back to Redshift DOUBLE PRECISION type, as Redshift "
                           "doesn't support fp_significand_bits>{FLOAT64_SIGNIFICAND_BITS} "
                           f"(requested: {fp_significand_bits}")
            return sqlalchemy.sql.sqltypes.Float(precision=FLOAT64_SIGNIFICAND_BITS)
        return super().type_for_floating_point(fp_total_bits=fp_total_bits,
                                               fp_significand_bits=fp_significand_bits)

    def loader(self) -> Optional[LoaderFromRecordsDirectory]:
        return self._redshift_loader

    def loader_from_fileobj(self) -> None:
        return None

    def unloader(self) -> Optional[Unloader]:
        return self._redshift_unloader

    def tweak_records_schema_for_load(self,
                                      records_schema: RecordsSchema,
                                      records_format: BaseRecordsFormat) -> RecordsSchema:
        if isinstance(records_format, AvroRecordsFormat):
            # upon testing, Redshift does not seem to support any of
            # Avro's logicalTypes - here's an example Avro schema; the
            # data will only import into the base type, not the
            # logical type :
            #
            # {"name": "date", "type": ["null", {"type": "int",
            #                                    "logicalType": "date" }]},
            # {"name": "time", "type": [ "null", {"type": "long",
            #                                     "logicalType": "time-micros"}]},
            # {"name": "timestamp", "type": [ "null", {"type": "string",
            #                                          "logicalType": "datetime"}]},
            # {"name": "timestamptz", "type": [ "null", {"type": "long",
            #                                            "logicalType": "timestamp-micros"}]}]}
            #
            # We could potentially tell Redshift to use
            # TIMEFORMAT/DATEFORMAT on import.  Unfortunately, while
            # it supports 'epochsecs' and 'epochmillis', it doesn't
            # support 'epochdays', or 'epochmicros', which would be
            # needed for the above.
            return records_schema.cast_field_types({
                'date': 'integer',
                'time': 'integer',
                'datetime': 'string',
                # timestamp-micros is too large for default integer size
                'datetimetz': 'string',
            })
        else:
            return records_schema



from urllib.parse import urlparse, unquote
from records_mover.url import BaseDirectoryUrl, BaseFileUrl
import google.auth.credentials
import google.cloud.storage
import googleapiclient.discovery
from typing import List, TYPE_CHECKING
if TYPE_CHECKING:
    from .gcs_file_url import GCSFileUrl


class GCSDirectoryUrl(BaseDirectoryUrl):
    def __init__(self,
                 url: str,
                 gcs_client: google.cloud.storage.Client,
                 gcp_credentials: google.auth.credentials.Credentials,
                 **kwargs) -> None:
        self.url = url
        parsed = urlparse(url)
        self.blob = unquote(parsed.path[1:])
        self.scheme = parsed.scheme
        self.bucket = parsed.netloc
        self.client = gcs_client
        self.credentials = gcp_credentials
        self.bucket_obj = self.client.bucket(self.bucket)

    def _directory(self, url: str) -> 'GCSDirectoryUrl':
        return GCSDirectoryUrl(url, gcs_client=self.client, gcp_credentials=self.credentials)

    def _file(self, url: str) -> 'GCSFileUrl':
        from .gcs_file_url import GCSFileUrl
        return GCSFileUrl(url, gcs_client=self.client, gcp_credentials=self.credentials)

    def directory_in_this_directory(self, directory_name: str) -> 'GCSDirectoryUrl':
        return self._directory(f"{self.url}{directory_name}/")

    def purge_directory(self) -> None:
        for dir_loc in self.directories_in_directory():
            dir_loc.purge_directory()
        for file_loc in self.files_in_directory():
            file_loc.delete()
        # Like S3, directories are just a convenient fiction in GCS;
        # once all contents are gone, they disappear from the file
        # listing, since they never really existed, so we don't need
        # a final delete on the directory.

    def files_in_directory(self) -> List[BaseFileUrl]:
        prefix = self.blob
        blobs = self.client.list_blobs(bucket_or_name=self.bucket,
                                       prefix=prefix, delimiter='/')
        blob_names = [
            blob.name
            for blob in blobs
            # I've seen this happen with gs://bluelabs-test-recordsmover/bar/
            if blob.name != prefix
        ]
        file_urls = [
            f"gs://{self.bucket}/{blob_name}"
            for blob_name in blob_names
        ]
        return [self._file(file_url) for file_url in file_urls]

    def directories_in_directory(self) -> List[BaseDirectoryUrl]:
        # https://stackoverflow.com/a/57099354/9795956

        service = googleapiclient.discovery.build('storage', 'v1',
                                                  credentials=self.credentials)
        prefix = self.blob
        folders_req = service.objects().list(bucket=self.bucket,
                                             prefix=prefix,
                                             delimiter='/')
        folders_resp = folders_req.execute()
        folder_names = folders_resp.get('prefixes', [])
        directory_urls = [
            f"gs://{self.bucket}/{folder_name}"
            for folder_name in folder_names
            # I haven't seen this happen - this is for safety
            if folder_name != prefix
        ]
        return [self._directory(directory_url) for directory_url in directory_urls]



import logging
from ..url.resolver import UrlResolver
from .sources import RecordsSources
from .targets import RecordsTargets
from .mover import move
from enum import Enum
from typing import Callable, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from sqlalchemy.engine import Engine, Connection  # noqa
    from ..db import DBDriver  # noqa

    from records_mover import Session  # noqa

logger = logging.getLogger(__name__)


class PleaseInfer(Enum):
    # This is a mypy-friendly way of doing a singleton object:
    #
    # https://github.com/python/typing/issues/236
    token = 1


class Records:
    """To move records from one place to another, you can use the methods on this object.

    This object should be pulled from the 'records' property on a
    :class:`records_mover.Session` object instead of being constructed
    directly.

    To move data, you can call the :meth:`records_mover.records.move`
    method, which is aliased for your convenience on this object.

    Example:

    .. code-block:: python

       records = session.records
       db_engine = session.get_default_db_engine()
       url = 's3://some-bucket/some-directory/'
       source = records.sources.directory_from_url(url=url)
       target = records.targets.table(schema_name='myschema',
                                      table_name='mytable',
                                      db_engine=db_engine)
       results = records.move(source, target)

    """

    sources: RecordsSources
    """Object containing factory methods to create various sources from
    which to copy records, of type :class:`records_mover.records.sources.RecordsSources`
    """

    targets: RecordsTargets
    """Object containing factory methods to create various targets to
which records can be copied, of type :class:`records_mover.records.targets.RecordsTargets`
    """

    move: Callable
    "Alias of :meth:`records_mover.records.move`"

    def __init__(self,
                 db_driver: Union[Callable[[Union['Engine', 'Connection']], 'DBDriver'],
                                  PleaseInfer] = PleaseInfer.token,
                 url_resolver: Union[UrlResolver, PleaseInfer] = PleaseInfer.token,
                 session: Union['Session', PleaseInfer] = PleaseInfer.token) -> None:
        if db_driver is PleaseInfer.token or url_resolver is PleaseInfer.token:
            if session is PleaseInfer.token:
                from records_mover import Session  # noqa

                session = Session()
            if db_driver is PleaseInfer.token:
                db_driver = session.db_driver
            if url_resolver is PleaseInfer.token:
                url_resolver = session.url_resolver
        self.move = move
        self.sources = RecordsSources(db_driver=db_driver,
                                      url_resolver=url_resolver)
        self.targets = RecordsTargets(url_resolver=url_resolver,
                                      db_driver=db_driver)



from typing import Optional
from .base import (SupportsMoveToRecordsDirectory,
                   SupportsToDataframesSource)
from ...db.quoting import quote_schema_and_table
from ...db import DBDriver
from ..records_directory import RecordsDirectory
from ..processing_instructions import ProcessingInstructions
from ..records_format import BaseRecordsFormat
from ..unload_plan import RecordsUnloadPlan
from ..results import MoveResult
from sqlalchemy.engine import Engine
from contextlib import contextmanager
from ..schema import RecordsSchema
from ...url.resolver import UrlResolver
from records_mover.url.base import BaseDirectoryUrl
import logging
from typing import Generator, Iterator, List, TYPE_CHECKING
if TYPE_CHECKING:
    from .dataframes import DataframesRecordsSource  # noqa
    from pandas import DataFrame  # noqa

logger = logging.getLogger(__name__)


class TableRecordsSource(SupportsMoveToRecordsDirectory,
                         SupportsToDataframesSource):
    records_format: Optional[BaseRecordsFormat]

    def __init__(self,
                 schema_name: str,
                 table_name: str,
                 driver: DBDriver,
                 url_resolver: UrlResolver) -> None:
        self.schema_name = schema_name
        self.table_name = table_name
        self.driver = driver
        self.unloader = driver.unloader()
        if self.unloader is not None:
            self.records_format = self.unloader.best_records_format()
        else:
            self.records_format = None
        self.url_resolver = url_resolver

    def known_supported_records_formats(self) -> List[BaseRecordsFormat]:
        unloader = self.driver.unloader()
        if unloader is None:
            return []
        return unloader.known_supported_records_formats_for_unload()

    def can_move_to_format(self,
                           target_records_format: BaseRecordsFormat) -> bool:
        unloader = self.driver.unloader()
        if unloader is None:
            return False
        return unloader.can_unload_format(target_records_format)

    def can_move_to_scheme(self, scheme: str) -> bool:
        unloader = self.driver.unloader()
        if unloader is None:
            # bulk export is not provided by this database
            logger.warning("No unloader configured for this database "
                           f"type ({self.driver.db_engine.name})")
            return False
        can_unload = unloader.can_unload_to_scheme(scheme)
        if not can_unload:
            logger.warning(f"This database ({self.driver.db_engine.name}) is "
                           f"not configured to export to {scheme}")
        return can_unload

    @contextmanager
    def to_dataframes_source(self,
                             processing_instructions: ProcessingInstructions) -> \
            Iterator['DataframesRecordsSource']:
        from .dataframes import DataframesRecordsSource  # noqa
        import pandas

        db = self.driver.db
        records_schema = self.pull_records_schema()

        if isinstance(db, Engine):
            connection = db.connect()
            columns = db.dialect.get_columns(connection, self.table_name, schema=self.schema_name)
            connection.close()
        else:
            columns = db.dialect.get_columns(db, self.table_name, schema=self.schema_name)

        num_columns = len(columns)
        if num_columns == 0:
            raise LookupError(f"Could not find {self.schema_name}.{self.table_name}")
        entries_per_chunk = 2000000
        chunksize = int(entries_per_chunk / num_columns)
        logger.info(f"Exporting in chunks of up to {chunksize} rows by {num_columns} columns")

        quoted_table = quote_schema_and_table(db, self.schema_name, self.table_name)
        chunks: Generator['DataFrame', None, None] = \
            pandas.read_sql(f"SELECT * FROM {quoted_table}",
                            con=db,
                            chunksize=chunksize)
        try:
            yield DataframesRecordsSource(dfs=self.with_cast_dataframe_types(records_schema,
                                                                             chunks),
                                          records_schema=records_schema,
                                          processing_instructions=processing_instructions)
        finally:
            chunks.close()

    def with_cast_dataframe_types(self,
                                  records_schema: RecordsSchema,
                                  dfs: Iterator['DataFrame']) -> Iterator['DataFrame']:
        for df in dfs:
            yield records_schema.cast_dataframe_types(df)

    def pull_records_schema(self) -> RecordsSchema:
        return RecordsSchema.from_db_table(self.schema_name, self.table_name,
                                           driver=self.driver)

    def move_to_records_directory(self,
                                  records_directory: RecordsDirectory,
                                  records_format: BaseRecordsFormat,
                                  processing_instructions: ProcessingInstructions,
                                  schema_name: Optional[str] = None,
                                  table_name: Optional[str] = None,
                                  engine: Optional[Engine] = None) -> MoveResult:
        unload_plan = RecordsUnloadPlan(records_format=records_format,
                                        processing_instructions=processing_instructions)
        unloader = self.driver.unloader()
        if unloader is None:
            raise ValueError('This DBDriver does not support bulk unloading')
        export_count = unloader.unload(schema=self.schema_name, table=self.table_name,
                                       unload_plan=unload_plan,
                                       directory=records_directory)
        records_schema = self.pull_records_schema()
        records_directory.save_format(unload_plan.records_format)
        records_schema = self.driver.tweak_records_schema_after_unload(records_schema,
                                                                       unload_plan.records_format)
        records_directory.save_schema(records_schema)
        records_directory.finalize_manifest()

        return MoveResult(move_count=export_count, output_urls=None)

    @contextmanager
    def temporary_unloadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        """Yield a temporary directory that can be used to call move_to_records_directory() on."""
        unloader = self.driver.unloader()
        if unloader is None:
            raise ValueError('This DBDriver does not support bulk unloading')
        with unloader.temporary_unloadable_directory_loc() as temp_loc:
            yield temp_loc

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.driver.db_engine.name})"



import logging
from sqlalchemy import Column
import sqlalchemy
from typing import cast, Callable, Union, Optional, Type, TYPE_CHECKING
from .constraints import (RecordsSchemaFieldConstraints,
                          RecordsSchemaFieldIntegerConstraints,
                          RecordsSchemaFieldDecimalConstraints,
                          RecordsSchemaFieldStringConstraints)
from .representation import RecordsSchemaFieldRepresentation
from .statistics import RecordsSchemaFieldStringStatistics
from .string_length_generator import generate_string_length
if TYPE_CHECKING:
    from ....db import DBDriver  # noqa
    from ..field import RecordsSchemaField  # noqa
    from ..schema import RecordsSchema  # noqa
    from .field_types import FieldType  # noqa


logger = logging.getLogger(__name__)


def field_from_sqlalchemy_column(column: Column,
                                 driver: 'DBDriver',
                                 rep_type: str)\
        -> 'RecordsSchemaField':
    from ..field import RecordsSchemaField  # noqa
    name = column.name
    type_: Union[Type[sqlalchemy.types.TypeEngine],
                 sqlalchemy.types.TypeEngine] = column.type
    constraints: Optional[RecordsSchemaFieldConstraints] = None
    field_type = None  # type: Optional[FieldType]

    required = not column.nullable

    # We don't yet try to pull information about uniqueness
    # constraints out of database tables, so set this to null,
    # which means "we don't know":
    #
    # https://github.com/bluelabsio/records-mover/issues/90
    unique = None

    if isinstance(type_, sqlalchemy.sql.sqltypes.Integer):
        field_type = 'integer'
        constraints = RecordsSchemaFieldIntegerConstraints.\
            from_sqlalchemy_type(required=required,
                                 unique=unique,
                                 type_=type_,
                                 driver=driver)
    elif (isinstance(type_, (sqlalchemy.sql.sqltypes.Numeric,
                             sqlalchemy.sql.sqltypes.Float))):
        field_type = 'decimal'
        constraints = RecordsSchemaFieldDecimalConstraints.\
            from_sqlalchemy_type(required=required,
                                 unique=unique,
                                 type_=type_,
                                 driver=driver)
    elif (isinstance(type_, sqlalchemy.sql.sqltypes.String)):
        field_type = 'string'
        constraints = RecordsSchemaFieldStringConstraints.\
            from_sqlalchemy_type(required=required,
                                 unique=unique,
                                 type_=type_,
                                 driver=driver)
    elif isinstance(type_, sqlalchemy.sql.sqltypes.Date):
        field_type = 'date'
    elif isinstance(type_, sqlalchemy.sql.sqltypes.DateTime):
        date_plus_time_with_timezone = driver.type_for_date_plus_time(has_tz=True)
        date_plus_time_no_timezone = driver.type_for_date_plus_time(has_tz=False)

        #
        # Databases are super annoying about these.
        #
        # BigQuery's SQLAlchemy driver uses TIMEZONE for the
        # former, DATETIME for the latter, without setting the
        # 'timezone' variable.
        #
        # Redshift and Vertica use DateTime for both with the
        # 'timezone' variable set appropriately.
        #
        if not isinstance(date_plus_time_with_timezone, type(date_plus_time_no_timezone)):
            if isinstance(type_, type(date_plus_time_with_timezone)):
                field_type = 'datetimetz'
            elif isinstance(type_, type(date_plus_time_no_timezone)):
                field_type = 'datetime'
            else:
                raise NotImplementedError(f"Teach me how to handle {type_}")
        else:
            if type_.timezone:
                field_type = 'datetimetz'
            else:
                field_type = 'datetime'
    elif isinstance(type_, sqlalchemy.sql.sqltypes.Time):
        if type_.timezone:
            field_type = 'timetz'
        else:
            field_type = 'time'
    elif isinstance(type_, sqlalchemy.sql.sqltypes.Boolean):
        field_type = 'boolean'
    else:
        raise NotImplementedError("Teach me how to handle this SQLAlchemy type: "
                                  f"{type(type_)}")
    if constraints is None:
        constraints = RecordsSchemaFieldConstraints.\
            from_sqlalchemy_type(required=required,
                                 unique=unique,
                                 type_=type_,
                                 driver=driver)
    representations = {
        'origin': RecordsSchemaFieldRepresentation.
        from_sqlalchemy_column(column, driver.db.dialect,
                               rep_type)
    }

    # We don't currently gather statistics from databases - which
    # can bite us when exporting from BigQuery, for instance:
    #
    # https://github.com/bluelabsio/records-mover/issues/91
    statistics = None

    return RecordsSchemaField(name=name,
                              field_type=field_type,
                              constraints=constraints,
                              statistics=statistics,
                              representations=representations)


def get_integer_type(field: 'RecordsSchemaField',
                     driver: 'DBDriver') -> sqlalchemy.types.TypeEngine:
    int_constraints = cast(Optional[RecordsSchemaFieldIntegerConstraints], field.constraints)
    min_ = getattr(int_constraints, 'min_', None)
    max_ = getattr(int_constraints, 'max_', None)
    if field.constraints and not isinstance(field.constraints,
                                            RecordsSchemaFieldIntegerConstraints):
        raise ValueError(f"Incorrect constraint type in {field.name}: {field.constraints}")
    return driver.type_for_integer(min_value=min_, max_value=max_)


def get_decimal_type(field: 'RecordsSchemaField',
                     driver: 'DBDriver') -> sqlalchemy.types.TypeEngine:
    decimal_constraints = cast(Optional[RecordsSchemaFieldDecimalConstraints],
                               field.constraints)
    fixed_precision = getattr(decimal_constraints, 'fixed_precision', None)
    fixed_scale = getattr(decimal_constraints, 'fixed_scale', None)
    fp_total_bits = getattr(decimal_constraints, 'fp_total_bits', None)
    fp_significand_bits = getattr(decimal_constraints, 'fp_significand_bits', None)
    if (fixed_precision and fixed_scale):
        return driver.type_for_fixed_point(precision=fixed_precision, scale=fixed_scale)
    elif (fp_total_bits and fp_significand_bits):
        return driver.type_for_floating_point(fp_total_bits=fp_total_bits,
                                              fp_significand_bits=fp_significand_bits)
    return driver.type_for_floating_point(fp_total_bits=64, fp_significand_bits=53)


def get_string_type(field: 'RecordsSchemaField', driver: 'DBDriver') -> sqlalchemy.types.TypeEngine:
    if field.constraints and not isinstance(field.constraints, RecordsSchemaFieldStringConstraints):
        raise ValueError(f"Incorrect constraint type in {field.name}: {field.constraints}")
    elif field.statistics and not isinstance(field.statistics, RecordsSchemaFieldStringStatistics):
        raise ValueError(f"Incorrect statistics type in {field.name}: {field.statistics}")
    return sqlalchemy.sql.sqltypes.String(generate_string_length(
        cast(Optional[RecordsSchemaFieldStringConstraints], field.constraints),
        cast(Optional[RecordsSchemaFieldStringStatistics], field.statistics),
        driver))


def field_to_sqlalchemy_type(field: 'RecordsSchemaField',
                             driver: 'DBDriver') -> sqlalchemy.types.TypeEngine:
    field_to_func_map = {
        'integer': get_integer_type,
        'decimal': get_decimal_type,
        'boolean': lambda field, driver: sqlalchemy.sql.sqltypes.BOOLEAN(),
        'string': get_string_type,
        'date': lambda field, driver: sqlalchemy.sql.sqltypes.DATE(),
        'datetime': lambda field, driver: driver.type_for_date_plus_time(has_tz=False),
        'datetimetz': lambda field, driver: driver.type_for_date_plus_time(has_tz=True),
        'time': lambda field, driver: (
            sqlalchemy.sql.sqltypes.TIME(timezone=False)
            if driver.supports_time_type()
            else sqlalchemy.sql.sqltypes.VARCHAR(8)),
        'timetz': lambda field, driver: (
            sqlalchemy.sql.sqltypes.TIME(timezone=True)
            if driver.supports_time_type()
            else sqlalchemy.sql.sqltypes.VARCHAR(8))
    }

    func = field_to_func_map.get(field.field_type, None)
    if not func:
        raise NotImplementedError(f"Teach me how to handle records schema type {field.field_type}")
    typed_func = cast(Callable[['RecordsSchemaField', 'DBDriver'], sqlalchemy.types.TypeEngine],
                      func)
    return typed_func(field, driver)


def field_to_sqlalchemy_column(field: 'RecordsSchemaField', driver: 'DBDriver') -> Column:
    name = field.name if driver is None else driver.make_column_name_valid(field.name)
    return Column(name,
                  type_=field.to_sqlalchemy_type(driver),
                  nullable=field.constraints is None or not field.constraints.required)



import logging
import boto3
from .s3_file_url import S3FileUrl
from .s3_directory_url import S3DirectoryUrl
from typing import Union

logger = logging.getLogger(__name__)


def S3Url(url: str, boto3_session: boto3.session.Session, **kwargs: object) -> Union[S3DirectoryUrl,
                                                                                     S3FileUrl]:
    if url.endswith('/'):
        return S3DirectoryUrl(url, boto3_session=boto3_session, S3Url=S3Url)
    else:
        return S3FileUrl(url, boto3_session=boto3_session, S3Url=S3Url)



from urllib.parse import urlparse, unquote
from records_mover.url.gcs.gcs_directory_url import GCSDirectoryUrl
from typing import IO
from records_mover.url import BaseFileUrl
from smart_open.gcs import open as gs_open
import google.cloud.storage
from google.cloud.storage.blob import Blob
import google.api_core.exceptions


class GCSFileUrl(BaseFileUrl):
    def __init__(self,
                 url: str,
                 gcs_client: google.cloud.storage.Client,
                 gcp_credentials: google.auth.credentials.Credentials,
                 **kwargs) -> None:
        self.url = url
        parsed = urlparse(url)
        self.blob = unquote(parsed.path[1:])
        self.bucket = parsed.netloc
        self.client = gcs_client
        self.credentials = gcp_credentials
        self.bucket_obj = self.client.bucket(self.bucket)

    def open(self, mode: str = "rb") -> IO[bytes]:
        try:
            return gs_open(bucket_id=self.bucket,
                           blob_id=self.blob,
                           mode=mode,
                           client=self.client)
        except google.api_core.exceptions.NotFound:
            # google.api_core.exceptions.NotFound: 404 blob
            # SkosHyDshbo/_manifest not found in bucket-name-here
            #
            # smart-open version: 2.0
            raise FileNotFoundError(f"{self} not found")

    def _directory(self, url: str) -> GCSDirectoryUrl:
        return GCSDirectoryUrl(url, gcs_client=self.client, gcp_credentials=self.credentials)

    def _blob_obj(self) -> 'Blob':
        return self.bucket_obj.blob(self.blob)

    def size(self) -> int:
        return self._blob_obj().size

    def rename_to(self, new: 'BaseFileUrl') -> 'BaseFileUrl':
        if not isinstance(new, GCSFileUrl):
            raise NotImplementedError('Cannot rename a GCS file to a non-GCS file')
        if new.bucket != self.bucket:
            raise NotImplementedError('Cannot rename a GCS file in a different bucket')
        # This actually does a copy behind the scenes, so don't get
        # too excited:
        self.bucket_obj.rename_blob(self._blob_obj(), new.blob)

        return new

    def filename(self) -> str:
        return self.url[self.url.rfind("/")+1:]

    def delete(self) -> None:
        self._blob_obj().delete()



from typing import Dict, Any, TypeVar, List, Union


V = TypeVar('V')


def nest_dict(d: Dict[str, V]) -> Dict[str, Union[V, Dict[str, Any]]]:
    """
    Takes a flat dict that has nested key names like this:

    {'abc': 1, 'foo.bar': 5, 'foo.baz.bing': 'bazzle'}

    ...and turns it into a nested dict that looks like this:

    {'abc': 1, 'foo': {'bar': 5, 'baz': {'bing': 'bazzle'}}}
    """
    def insert_into_dict(msg: Dict[str, Union[V, Dict[str, Any]]],
                         keys: List[str],
                         value: V) -> None:
        if len(keys) == 1:
            msg[keys[0]] = value
        else:
            if keys[0] not in msg:
                msg[keys[0]] = {}
            assert isinstance(msg[keys[0]], dict)
            insert_into_dict(msg[keys[0]], keys[1:], value)  # type: ignore
    msg: Dict[str, Union[V, Dict[str, Any]]] = {}
    for key, val in d.items():
        insert_into_dict(msg, key.split('.'), val)
    return msg



__all__ = [
    'RecordsSchemaKnownRepresentation',
    'RecordsSchema',
    'DeserializationError',
    'UnsupportedSchemaError',
    'RecordsSchemaField',
    'RecordsSchemaFieldConstraints',
    'RecordsSchemaFieldStatistics',
    'RecordsSchemaFieldRepresentation',
]

from .errors import DeserializationError, UnsupportedSchemaError
from .schema.known_representation import RecordsSchemaKnownRepresentation
from .schema import RecordsSchema
from .field import RecordsSchemaField
from .field.constraints import RecordsSchemaFieldConstraints
from .field.statistics import RecordsSchemaFieldStatistics
from .field.representation import RecordsSchemaFieldRepresentation



import logging
import json
from ...processing_instructions import ProcessingInstructions
from abc import ABCMeta, abstractmethod
from typing import Dict, Any, cast, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ....db import DBDriver  # noqa
    from mypy_extensions import TypedDict
    from pandas import DataFrame

    class KnownRepresentationDict(TypedDict):
        type: str

    class SqlKnownRepresentationDict(KnownRepresentationDict, total=False):
        table_ddl: str

    class CorePandasDataframeKnownRepresentationDict(KnownRepresentationDict):
        pd_df_dtypes: Dict[str, Dict[str, Any]]

    class PandasDataframeKnownRepresentationDict(CorePandasDataframeKnownRepresentationDict,
                                                 total=False):
        pd_df_ftypes: Dict[str, str]


logger = logging.getLogger(__name__)


class RecordsSchemaKnownRepresentation(metaclass=ABCMeta):

    @staticmethod
    def from_data(data: 'KnownRepresentationDict') -> Optional['RecordsSchemaKnownRepresentation']:
        if data['type'].startswith('sql/'):
            sql_data = cast('SqlKnownRepresentationDict', data)
            return RecordsSchemaSqlKnownRepresentation(type=sql_data['type'],
                                                       table_ddl=sql_data.get('table_ddl'))
        elif data['type'].startswith('dataframe/pandas'):
            pandas_data = cast('PandasDataframeKnownRepresentationDict', data)
            return RecordsSchemaPandasDataframeKnownRepresentation(
                pd_df_dtypes=pandas_data['pd_df_dtypes'],
                pd_df_ftypes=pandas_data.get('pd_df_ftypes'))
        else:
            logger.warning(f"Unknown known-representation type ({data['type']})--"
                           "consider upgrading records-mover")
            return None

    @staticmethod
    def from_dataframe(df: 'DataFrame',
                       processing_instructions: ProcessingInstructions) ->\
            'RecordsSchemaKnownRepresentation':
        dtypes_json = df.dtypes.to_json(default_handler=str)
        dtypes_data = json.loads(dtypes_json)
        ftypes_data = None
        if getattr(df, 'ftypes', None) is not None:
            ftypes_json = df.ftypes.to_json()
            ftypes_data = json.loads(ftypes_json)
        return RecordsSchemaPandasDataframeKnownRepresentation(pd_df_dtypes=dtypes_data,
                                                               pd_df_ftypes=ftypes_data)

    @staticmethod
    def from_db_driver(driver: 'DBDriver',
                       schema_name: str,
                       table_name: str) -> 'RecordsSchemaSqlKnownRepresentation':
        type = f"sql/{driver.db.dialect.name}"
        ddl = driver.schema_sql(schema_name, table_name)
        return RecordsSchemaSqlKnownRepresentation(type=type, table_ddl=ddl)

    @abstractmethod
    def to_data(self) -> 'KnownRepresentationDict':
        raise NotImplementedError


class RecordsSchemaSqlKnownRepresentation(RecordsSchemaKnownRepresentation):
    def __init__(self, type: str, table_ddl: Optional[str]) -> None:
        self.type = type
        self.table_ddl = table_ddl

    def to_data(self) -> 'SqlKnownRepresentationDict':
        out: 'SqlKnownRepresentationDict' = {
            'type': self.type
        }
        if self.table_ddl is not None:
            out['table_ddl'] = self.table_ddl
        return out


class RecordsSchemaPandasDataframeKnownRepresentation(RecordsSchemaKnownRepresentation):
    def __init__(self,
                 pd_df_dtypes: Dict[str, Dict[str, Any]],
                 pd_df_ftypes: Optional[Dict[str, str]] = None) -> None:
        self.pd_df_dtypes = pd_df_dtypes
        self.pd_df_ftypes = pd_df_ftypes

    def to_data(self) -> 'PandasDataframeKnownRepresentationDict':
        out: 'PandasDataframeKnownRepresentationDict' = {
            'type': 'dataframe/pandas',
            'pd_df_dtypes': self.pd_df_dtypes,
        }
        if self.pd_df_ftypes is not None:
            out['pd_df_ftypes'] = self.pd_df_ftypes
        return out



from records_mover.utils import quiet_remove
from records_mover.records.delimited import cant_handle_hint, ValidatedRecordsHints
from typing import Set
from .mode import CopyOptionsMode
from .types import PostgresCopyOptions, CopyOptionsModeType, _assert_never
from .common import postgres_copy_options_common


def postgres_copy_options_text(unhandled_hints: Set[str],
                               hints: ValidatedRecordsHints,
                               fail_if_cant_handle_hint: bool,
                               mode: CopyOptionsModeType) ->\
        PostgresCopyOptions:
    postgres_options: PostgresCopyOptions = {}

    # FORMAT
    #
    #  Selects the data format to be read or written: text, csv (Comma
    #  Separated Values), or binary. The default is text.
    #
    postgres_options['format'] = 'text'

    # Backslash characters (\) can be used in the COPY data to quote
    # data characters that might otherwise be taken as row or column
    # delimiters. In particular, the following characters must be
    # preceded by a backslash if they appear as part of a column
    # value: backslash itself, newline, carriage return, and the
    # current delimiter character.
    if hints.escape is None:
        cant_handle_hint(fail_if_cant_handle_hint, 'escape', hints)
    else:
        quiet_remove(unhandled_hints, 'escape')

    # NULL
    #
    #  Specifies the string that represents a null value. The default
    #  is \N (backslash-N) in text format, and an unquoted empty
    #  string in CSV format. You might prefer an empty string even in
    #  text format for cases where you don't want to distinguish nulls
    #  from empty strings. This option is not allowed when using
    #  binary format.
    #
    #  Note: When using COPY FROM, any data item that matches this
    #  string will be stored as a null value, so you should make sure
    #  that you use the same string as you used with COPY TO.
    #

    # QUOTE
    #
    #  Specifies the quoting character to be used when a data value is
    #  quoted. The default is double-quote. This must be a single
    #  one-byte character. This option is allowed only when using CSV
    #  format.
    #

    quiet_remove(unhandled_hints, 'quotechar')

    # ESCAPE
    #
    #  Specifies the character that should appear before a data
    #  character that matches the QUOTE value. The default is the same
    #  as the QUOTE value (so that the quoting character is doubled if
    #  it appears in the data). This must be a single one-byte
    #  character. This option is allowed only when using CSV format.
    #

    if hints.doublequote:
        cant_handle_hint(fail_if_cant_handle_hint, 'doublequote', hints)
    else:
        quiet_remove(unhandled_hints, 'doublequote')

    # FORCE_QUOTE
    #
    #  Forces quoting to be used for all non-NULL values in each
    #  specified column. NULL output is never quoted. If * is
    #  specified, non-NULL values will be quoted in all columns. This
    #  option is allowed only in COPY TO, and only when using CSV
    #  format.
    #

    if hints.quoting is not None:
        cant_handle_hint(fail_if_cant_handle_hint, 'quoting', hints)
    else:
        quiet_remove(unhandled_hints, 'quoting')

    # FORCE_NOT_NULL
    #
    #  Do not match the specified columns' values against the null
    #  string. In the default case where the null string is empty,
    #  this means that empty values will be read as zero-length
    #  strings rather than nulls, even when they are not quoted. This
    #  option is allowed only in COPY FROM, and only when using CSV
    #  format.
    #

    # COPY TO will terminate each row with a Unix-style newline
    # ("\n"). Servers running on Microsoft Windows instead output
    # carriage return/newline ("\r\n"), but only for COPY to a server
    # file; for consistency across platforms, COPY TO STDOUT always
    # sends "\n" regardless of server platform. COPY FROM can handle
    # lines ending with newlines, carriage returns, or carriage
    # return/newlines. To reduce the risk of error due to
    # un-backslashed newlines or carriage returns that were meant as
    # data, COPY FROM will complain if the line endings in the input
    # are not all alike.

    if mode is CopyOptionsMode.LOADING:
        if hints.record_terminator in ["\n", "\r", "\r\n"]:
            quiet_remove(unhandled_hints, 'record-terminator')
        else:
            cant_handle_hint(fail_if_cant_handle_hint, 'record-terminator', hints)
    elif mode is CopyOptionsMode.UNLOADING:
        # No control for this is given - exports appear with unix
        # newlines.
        if hints.record_terminator == "\n":
            quiet_remove(unhandled_hints, 'record-terminator')
        else:
            cant_handle_hint(fail_if_cant_handle_hint, 'record-terminator', hints)
    else:
        _assert_never(mode)

    if hints.compression is not None:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', hints)
    else:
        quiet_remove(unhandled_hints, 'compression')

    return postgres_copy_options_common(unhandled_hints,
                                        hints,
                                        fail_if_cant_handle_hint,
                                        postgres_options)



from contextlib import contextmanager
from typing import IO, Iterator
import logging


logger = logging.getLogger(__name__)


@contextmanager
def rewound_fileobj(fileobj: IO[bytes]) -> Iterator[IO[bytes]]:
    if getattr(fileobj, 'closed', None) is not None:
        closed = fileobj.closed
    if closed:
        logger.warning("Stream already closed")
        raise OSError('Stream is already closed')
    if not fileobj.seekable():
        # OSError is what is thrown when you call .seek() on a
        # non-rewindable stream.
        raise OSError('Stream is not rewindable')
    original_position = fileobj.tell()
    fileobj.seek(0)
    try:
        yield fileobj
    finally:
        fileobj.seek(original_position)



from .driver import GenericDBDriver, DBDriver
import sqlalchemy
from typing import Union


def db_driver(db: Union[sqlalchemy.engine.Engine,
                        sqlalchemy.engine.Connection],
              **kwargs) -> DBDriver:
    engine: sqlalchemy.engine.Engine = db.engine
    engine_name: str = engine.name

    if engine_name == 'vertica':
        from .vertica.vertica_db_driver import VerticaDBDriver

        return VerticaDBDriver(db, **kwargs)
    elif engine_name == 'redshift':
        from .redshift.redshift_db_driver import RedshiftDBDriver

        return RedshiftDBDriver(db, **kwargs)
    elif engine_name == 'bigquery':
        from .bigquery.bigquery_db_driver import BigQueryDBDriver

        return BigQueryDBDriver(db, **kwargs)
    elif engine_name == 'postgresql':
        from .postgres.postgres_db_driver import PostgresDBDriver

        return PostgresDBDriver(db, **kwargs)
    elif engine_name == 'mysql':
        from .mysql.mysql_db_driver import MySQLDBDriver

        return MySQLDBDriver(db, **kwargs)
    else:
        return GenericDBDriver(db, **kwargs)



import logging
from typing import Optional, Union, cast, TYPE_CHECKING
if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    from .field_types import FieldType  # noqa

    class FieldStatisticsDict(TypedDict):
        rows_sampled: int
        total_rows: int

    class StringFieldStatisticsDict(FieldStatisticsDict, total=False):
        max_length_bytes: int
        max_length_chars: int


logger = logging.getLogger(__name__)


class RecordsSchemaFieldStatistics:
    def __init__(self,
                 rows_sampled: int,
                 total_rows: int):
        self.rows_sampled = rows_sampled
        self.total_rows = total_rows

    def to_data(self) -> 'FieldStatisticsDict':
        return {
            'rows_sampled': self.rows_sampled,
            'total_rows': self.total_rows
        }

    @staticmethod
    def from_data(data: Optional[Union['FieldStatisticsDict', 'StringFieldStatisticsDict']],
                  field_type: 'FieldType') -> Optional['RecordsSchemaFieldStatistics']:
        if data is None:
            return None
        rows_sampled = data['rows_sampled']
        total_rows = data['total_rows']
        if field_type == 'string':
            data = cast('StringFieldStatisticsDict', data)
            return RecordsSchemaFieldStringStatistics(
                rows_sampled=rows_sampled,
                total_rows=total_rows,
                max_length_bytes=data.get('max_length_bytes'),
                max_length_chars=data.get('max_length_chars'))
        else:
            return RecordsSchemaFieldStatistics(rows_sampled=rows_sampled,
                                                total_rows=total_rows)

    def cast(self, field_type: 'FieldType') -> Optional['RecordsSchemaFieldStatistics']:
        # only string provides statistics at this point
        return None

    def __str__(self) -> str:
        return f"{type(self)}({self.to_data()})"


class RecordsSchemaFieldStringStatistics(RecordsSchemaFieldStatistics):
    def __init__(self,
                 rows_sampled: int,
                 total_rows: int,
                 max_length_bytes: Optional[int],
                 max_length_chars: Optional[int]):
        super().__init__(rows_sampled=rows_sampled,
                         total_rows=total_rows)
        self.max_length_bytes = max_length_bytes
        self.max_length_chars = max_length_chars

    def to_data(self) -> 'StringFieldStatisticsDict':
        generic_out = super().to_data()
        out = cast('StringFieldStatisticsDict', generic_out)
        if self.max_length_bytes is not None:
            out['max_length_bytes'] = int(self.max_length_bytes)
        if self.max_length_chars is not None:
            out['max_length_chars'] = int(self.max_length_chars)
        return out

    def merge(self, other: 'RecordsSchemaFieldStringStatistics') -> None:
        raise NotImplementedError

    def cast(self, field_type: 'FieldType') -> Optional['RecordsSchemaFieldStatistics']:
        if field_type == 'string':
            return self
        else:
            return super().cast(field_type)



__all__ = [
    'UntypedRecordsHints',
    'validate_partial_hints',
    'cant_handle_hint',
    'complain_on_unhandled_hints',
    'PartialRecordsHints',
    'ValidatedRecordsHints',
    'sniff_compression_from_url',
    'HintEncoding',
    'HintRecordTerminator',
    'HintFieldDelimiter',
    'HintQuoteChar',
    'HintQuoting',
    'HintEscape',
    'HintHeaderRow',
    'HintCompression',
    'HintDoublequote',
    'sniff_hints_from_fileobjs',
    'python_encoding_from_hint',
    'python_date_format_from_hints',
    'python_time_format_from_hints',
    'stream_csv',
]

"""RecordsHints are described as part of the overall `records format
documentation
<https://github.com/bluelabsio/knowledge/blob/master/Engineering/Architecture/JobDataExchange/output-design.md#hints>`_
Their goal in life is to both describe in detail how to write a
records file (especially delimited files, which are particularly
thorny in practice).

See the :py:meth:`records_mover.records.records_format.RecordsFormat`
for the other details that are typically provided along with records
hints.
"""

from .types import PartialRecordsHints, UntypedRecordsHints
from .validated_records_hints import ValidatedRecordsHints
from .hints import validate_partial_hints
from .utils import cant_handle_hint, complain_on_unhandled_hints
from .compression import sniff_compression_from_url
from .types import (
    HintEncoding, HintRecordTerminator,
    HintFieldDelimiter, HintQuoteChar,
    HintQuoting, HintEscape,
    HintHeaderRow, HintCompression,
    HintDoublequote,
)
from .sniff import sniff_hints_from_fileobjs
from .csv_streamer import stream_csv
from .conversions import (
    python_encoding_from_hint, python_date_format_from_hints, python_time_format_from_hints
)



from .s3_base_url import S3BaseUrl
from .awscli import aws_cli
from ..base import BaseDirectoryUrl, BaseFileUrl
from ..filesystem import FilesystemDirectoryUrl
from typing import Dict, List


class S3DirectoryUrl(S3BaseUrl, BaseDirectoryUrl):
    def directory_in_this_directory(self, directory_name: str) -> 'S3DirectoryUrl':
        return self._directory(f"{self.url}{directory_name}/")

    def files_in_directory(self) -> List['BaseFileUrl']:
        prefix = self.key
        resp = self.s3_client.list_objects(Bucket=self.bucket, Prefix=prefix, Delimiter='/')
        keys = [item['Key'] for item in resp.get('Contents', [])]
        return [self._key_in_same_bucket(key) for key in keys]

    def directories_in_directory(self) -> List['BaseDirectoryUrl']:
        prefix = self.key
        response = self.s3_client.list_objects_v2(Bucket=self.bucket, Prefix=prefix, Delimiter='/')
        prefix_keys = [content['Prefix'] for content in response.get('CommonPrefixes', [])]
        urls = [f"s3://{self.bucket}/{prefix}{key}" for key in prefix_keys]
        return [self._directory(url) for url in urls]

    def purge_directory(self) -> None:
        if not self.is_directory():
            raise ValueError("Not a directory")
        # https://stackoverflow.com/questions/11426560/amazon-s3-boto-how-to-delete-folder
        objects_to_delete = self.s3_resource.meta.client.list_objects(Bucket=self.bucket,
                                                                      Prefix=self.key)
        delete_keys: Dict[str, List[Dict[str, str]]] = {'Objects': []}
        delete_keys['Objects'] =\
            [{'Key': k} for k in [obj['Key'] for obj in objects_to_delete.get('Contents', [])]]
        if delete_keys['Objects']:
            self.s3_resource.meta.client.delete_objects(Bucket=self.bucket, Delete=delete_keys)

    def copy_to(self, other_loc: BaseDirectoryUrl) -> BaseDirectoryUrl:
        if not other_loc.is_directory():
            raise RuntimeError(f"Cannot copy a directory to a file ({other_loc.url})")
        elif isinstance(other_loc, FilesystemDirectoryUrl):
            aws_cli('s3', 'sync', self.url, other_loc.local_file_path)
            return other_loc
        else:
            return super(S3DirectoryUrl, self).copy_to(other_loc)

    def files_matching_prefix(self, prefix: str) -> List[BaseFileUrl]:
        prefix = self.key + prefix
        resp = self.s3_client.list_objects(Bucket=self.bucket, Prefix=prefix, Delimiter='/')
        return [self._key_in_same_bucket(o['Key']) for o in resp.get('Contents', [])]



from abc import ABCMeta, abstractmethod
import json
from typing import Mapping, Optional
from .records_types import RecordsFormatType


class BaseRecordsFormat(metaclass=ABCMeta):
    """This represents the information needed to be able to parse a set of
    records data files.

    See the `records format specification
    <https://github.com/bluelabsio/records-mover/blob/master/docs/RECORDS_SPEC.md>`_
    for more detail.

    To create an instance, see
    :class:`~records_mover.records.ParquetRecordsFormat` or
    :class:`~records_mover.records.DelimitedRecordsFormat`
    """
    format_type: RecordsFormatType

    def json(self) -> str:
        contents = ''
        contents_data = self.config()
        if contents_data is not None:
            contents = json.dumps(contents_data)
        return contents

    def config(self) -> Optional[Mapping[str, object]]:
        return None

    @abstractmethod
    def generate_filename(self, basename: str) -> str:
        pass



import sqlalchemy
import logging
from records_mover.url.resolver import UrlResolver
from records_mover.utils.limits import (INT16_MIN, INT16_MAX,
                                        INT32_MIN, INT32_MAX,
                                        INT64_MIN, INT64_MAX,
                                        FLOAT32_SIGNIFICAND_BITS,
                                        FLOAT64_SIGNIFICAND_BITS,
                                        num_digits)
from ..driver import DBDriver
from .loader import PostgresLoader
from ..loader import LoaderFromFileobj, LoaderFromRecordsDirectory
from .unloader import PostgresUnloader
from ..unloader import Unloader
from typing import Optional, Tuple, Union


logger = logging.getLogger(__name__)


class PostgresDBDriver(DBDriver):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
                 url_resolver: UrlResolver,
                 **kwargs) -> None:
        super().__init__(db)
        self._postgres_loader = PostgresLoader(url_resolver=url_resolver,
                                               meta=self.meta,
                                               db=self.db)
        self._postgres_unloader = PostgresUnloader(db=self.db)

    def loader(self) -> Optional[LoaderFromRecordsDirectory]:
        return self._postgres_loader

    def loader_from_fileobj(self) -> LoaderFromFileobj:
        return self._postgres_loader

    def unloader(self) -> Optional[Unloader]:
        return self._postgres_unloader

    # https://www.postgresql.org/docs/10/datatype-numeric.html
    def integer_limits(self,
                       type_: sqlalchemy.types.Integer) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.sql.sqltypes.SMALLINT):
            return (INT16_MIN, INT16_MAX)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.INTEGER):
            return (INT32_MIN, INT32_MAX)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.BIGINT):
            return (INT64_MIN, INT64_MAX)
        return super().integer_limits(type_)

    def fp_constraints(self,
                       type_: sqlalchemy.types.Float) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.dialects.postgresql.base.DOUBLE_PRECISION):
            return (64, FLOAT64_SIGNIFICAND_BITS)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.REAL):
            return (32, FLOAT32_SIGNIFICAND_BITS)
        return super().fp_constraints(type_)

    def type_for_integer(self,
                         min_value: Optional[int],
                         max_value: Optional[int]) -> sqlalchemy.types.TypeEngine:
        """Find correct integral column type to fit the given min and max integer values"""

        if min_value is not None and max_value is not None:
            if min_value >= INT16_MIN and max_value <= INT16_MAX:
                return sqlalchemy.sql.sqltypes.SMALLINT()
            if min_value >= INT32_MIN and max_value <= INT32_MAX:
                return sqlalchemy.sql.sqltypes.INTEGER()
            if min_value >= INT64_MIN and max_value <= INT64_MAX:
                return sqlalchemy.sql.sqltypes.BIGINT()
            else:
                num_digits_min = num_digits(min_value)
                num_digits_max = num_digits(max_value)
                digit_count = max(num_digits_min, num_digits_max)
                return self.type_for_fixed_point(precision=digit_count,
                                                 scale=0)
        return super().type_for_integer(min_value, max_value)

    def type_for_floating_point(self,
                                fp_total_bits: int,
                                fp_significand_bits: int) -> sqlalchemy.sql.sqltypes.Numeric:
        if fp_significand_bits > FLOAT64_SIGNIFICAND_BITS:
            logger.warning(f"Falling back to Postgres DOUBLE PRECISION type, as Postgres "
                           "doesn't support fp_significand_bits>{FLOAT64_SIGNIFICAND_BITS} "
                           f"(requested: {fp_significand_bits}")
            return sqlalchemy.sql.sqltypes.Float(precision=FLOAT64_SIGNIFICAND_BITS)
        return super().type_for_floating_point(fp_total_bits=fp_total_bits,
                                               fp_significand_bits=fp_significand_bits)



from .fileobjs import FileobjsSource
from .base import SupportsToFileobjsSource
from contextlib import contextmanager
from ..schema import RecordsSchema
from ..processing_instructions import ProcessingInstructions
from ..records_format import BaseRecordsFormat
from records_mover.records.delimited import (
    UntypedRecordsHints, validate_partial_hints
)
import logging
from typing import Optional, Iterator, Mapping, IO

logger = logging.getLogger(__name__)


class UninferredFileobjsRecordsSource(SupportsToFileobjsSource):
    def __init__(self,
                 target_names_to_input_fileobjs: Mapping[str, IO[bytes]],
                 records_format: Optional[BaseRecordsFormat] = None,
                 records_schema: Optional[RecordsSchema] = None,
                 initial_hints: Optional[UntypedRecordsHints] = None) -> None:
        self.target_names_to_input_fileobjs = target_names_to_input_fileobjs
        self.records_format = records_format
        self.records_schema = records_schema
        self.initial_hints = initial_hints
        if self.initial_hints is None:
            self.initial_hints = {}

    @contextmanager
    def to_fileobjs_source(self,
                           processing_instructions: ProcessingInstructions,
                           records_format_if_possible: Optional[BaseRecordsFormat] = None)\
            -> Iterator['FileobjsSource']:
        """Convert current source to a FileObjsSource and present it in a context manager"""
        typed_hints = None
        if self.initial_hints is not None:
            typed_hints =\
                validate_partial_hints(self.initial_hints,
                                       fail_if_cant_handle_hint=processing_instructions.
                                       fail_if_cant_handle_hint)
        with FileobjsSource.\
                infer_if_needed(target_names_to_input_fileobjs=self.target_names_to_input_fileobjs,
                                records_format=self.records_format,
                                records_schema=self.records_schema,
                                processing_instructions=processing_instructions,
                                initial_hints=typed_hints) as fileobjs_source:
            yield fileobjs_source



from pandas import DataFrame
from .known_representation import RecordsSchemaKnownRepresentation
from typing import Dict, TYPE_CHECKING
from ...processing_instructions import ProcessingInstructions
if TYPE_CHECKING:
    from ..field import RecordsSchemaField  # noqa
    from ..schema import RecordsSchema  # noqa


def schema_from_dataframe(df: DataFrame,
                          processing_instructions: ProcessingInstructions,
                          include_index: bool) -> 'RecordsSchema':
    from records_mover.records.schema import RecordsSchema  # noqa
    from records_mover.records.schema.field import RecordsSchemaField  # noqa
    fields = []
    origin_representation = \
        RecordsSchemaKnownRepresentation.from_dataframe(df, processing_instructions)
    known_representations: Dict[str, RecordsSchemaKnownRepresentation] = {
        'origin': origin_representation
    }

    if include_index:
        fields.append(RecordsSchemaField.
                      from_index(df.index,
                                 processing_instructions=processing_instructions))
    for column in df:
        fields.append(RecordsSchemaField.
                      from_series(df[column],
                                  processing_instructions=processing_instructions))

    return RecordsSchema(fields=fields,
                         known_representations=known_representations)


def refine_schema_from_dataframe(records_schema: 'RecordsSchema',
                                 df: DataFrame,
                                 processing_instructions:
                                 ProcessingInstructions = ProcessingInstructions()) ->\
        'RecordsSchema':
    from records_mover.records.schema import RecordsSchema

    max_sample_size = processing_instructions.max_inference_rows
    total_rows = len(df.index)
    if max_sample_size is not None and max_sample_size < total_rows:
        sampled_df = df.sample(n=max_sample_size)
    else:
        sampled_df = df
    rows_sampled = len(sampled_df.index)

    fields = [
        field.refine_from_series(sampled_df[field.name],
                                 total_rows=total_rows,
                                 rows_sampled=rows_sampled)
        for field in records_schema.fields
    ]
    return RecordsSchema(fields=fields,
                         known_representations=records_schema.known_representations)



import argparse
from records_mover.utils.structures import nest_dict
import os
import json
from ..mover_types import JsonSchema, JobConfig
from typing import Iterable, Dict, Any, Sequence, List, TypeVar, TYPE_CHECKING
if TYPE_CHECKING:
    from argparse_types import ArgParseArgument

ARG_VALUE_WAS_NONE: object = object()


A = TypeVar('A')


def arguments_output_to_config(arguments_output: Dict[str, A]) -> JobConfig:
    """
    argparse represents unspecified arguents as None, but we want to be
    able to know the difference between 'user told us None' (key set
    in dictionary and set to None) and 'not specified' (key not set in
    dict).
    """
    clean_config = {
        k: None if v == ARG_VALUE_WAS_NONE else v
        for k, v in arguments_output.items()
        if v is not None
    }
    return nest_dict(clean_config)


class JobConfigSchemaAsArgsParser():

    @staticmethod
    def from_description(config_json_schema: JsonSchema,
                         description: str) -> 'JobConfigSchemaAsArgsParser':
        argument_parser =\
            argparse.ArgumentParser(description=description,
                                    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        return JobConfigSchemaAsArgsParser(config_json_schema=config_json_schema,
                                           argument_parser=argument_parser)

    def __init__(self, config_json_schema: JsonSchema,
                 argument_parser: argparse.ArgumentParser) -> None:
        self.config_json_schema = config_json_schema
        self.arg_parser = argument_parser

    def add_to_prefix(self, prefix: str, key: str) -> str:
        """Return the period-separated representation of a nested JSON object.

        This allows nested JSON objects to be specified on the command
        line by representing them as a single parameter.

        (e.g., ./foo --a.b=c maps to {"a": { "b": "c"}}
        """
        if prefix == '':
            return key
        return prefix + '.' + key

    def configure_from_properties(self,
                                  properties: JsonSchema,
                                  required_keys: Iterable[str],
                                  prefix: str = '') -> None:
        for naked_key, raw_value in properties.items():
            if not isinstance(raw_value, dict):
                raise TypeError(f"Did not understand [{raw_value}] in [{properties}]")
            value: Dict[str, object] = raw_value
            key = self.add_to_prefix(prefix, naked_key)
            self.configure_from_single_property(key, value, required_keys, properties)

    def configure_from_single_property(self,
                                       key: str,
                                       value: Dict[str, object],
                                       required_keys: Iterable[str],
                                       properties: JsonSchema) -> None:
        arg_name = key
        # For consistency with other args, we map dashes to underscores in the CLI args
        formatted_key_name = key.replace('-', '_')
        kwargs: ArgParseArgument = {}
        if key in required_keys:
            # use a positional argument if it's required
            arg_name = key
        else:
            # use a --parameter=value argument if it's optional
            arg_name = "--" + formatted_key_name
            kwargs['required'] = False

            # When the formatted_key_name differs from key, this will make sure that the arg
            # parser is storing the value in a place that matches the actual schema
            kwargs['dest'] = key

        if 'default' in value:
            kwargs['default'] = value['default']

        if 'description' in value:
            kwargs['help'] = str(value['description'])

        if 'enum' in value:
            enum_values: Iterable[Any] = value['enum']  # type: ignore
            non_none_values: Iterable[Any] = [v for v in enum_values if v is not None]
            kwargs['choices'] = non_none_values
            if None in enum_values:
                # add an arg that maps to 'None' for this key
                no_arg_kwargs: ArgParseArgument = {}
                # if --no_foo is specified, set foo variable to None
                split_key_name = formatted_key_name.split('.')
                no_arg_arg_name = '--' + '.'.join([*split_key_name[:-1],
                                                   'no_' + split_key_name[-1]])
                no_arg_kwargs['action'] = 'store_const'
                no_arg_kwargs['const'] = ARG_VALUE_WAS_NONE
                no_arg_kwargs['dest'] = key
                no_arg_kwargs['required'] = False
                self.arg_parser.add_argument(no_arg_arg_name, **no_arg_kwargs)

        elif 'type' in value and value['type'] == 'string':
            kwargs['type'] = str
        elif 'type' in value and value['type'] == 'integer':
            kwargs['type'] = int
        elif 'type' in value and value['type'] == 'array':
            items = value['items']
            if not isinstance(items, dict):
                raise TypeError(f"Did not understand [{items}] in [{properties}]")
            item_type = items['type']
            if not isinstance(item_type, str):
                raise TypeError(f"Did not understand [{item_type}] in [{properties}]")
            if item_type == 'string':
                kwargs['type'] = str
            elif item_type == 'integer':
                kwargs['type'] = int
            else:
                raise NotImplementedError("Teach me how to handle array item "
                                          f"type of {item_type}")
            kwargs['nargs'] = '*'
        elif 'type' in value and value['type'] == 'object':
            # Recurse and handle nested dicts
            sub_properties = value['properties']
            if not isinstance(sub_properties, dict):
                raise TypeError(f"Did not understand [{sub_properties}] in [{properties}]")
            required_subkeys: List[str] = []
            if key in required_keys:
                required_subkeys = [
                    self.add_to_prefix(key, partial_subkey)
                    for partial_subkey
                    in value.get('required', [])  # type: ignore
                ]
            self.configure_from_properties(sub_properties, required_subkeys, prefix=key)
            return  # no immediate argument to add
        elif 'type' in value and value['type'] == 'boolean':
            # https://stackoverflow.com/questions/9183936/boolean-argument-for-script
            if 'default' in value:
                if value['default'] is True:
                    # if --no_foo is specified, set foo variable to False
                    arg_name = "--no_" + formatted_key_name
                    kwargs['action'] = 'store_false'
                else:
                    kwargs['action'] = 'store_true'
            else:
                if key in required_keys:
                    raise NotImplementedError("Teach me how to handle a required boolean "
                                              "with no default")
                else:
                    # the regular --foo will set this to True:
                    kwargs['action'] = 'store_const'
                    # store_true sets a default, so we use store_const
                    kwargs['const'] = True

                    # and we'll add an arg that maps to False for this key:
                    false_arg_kwargs: ArgParseArgument = {}
                    # if --no_foo is specified, set foo variable to False
                    split_key_name = formatted_key_name.split('.')
                    false_arg_arg_name = '--' + '.'.join([*split_key_name[:-1],
                                                          'no_' + split_key_name[-1]])
                    false_arg_kwargs['action'] = 'store_const'
                    false_arg_kwargs['const'] = False
                    false_arg_kwargs['dest'] = key
                    false_arg_kwargs['required'] = False
                    self.arg_parser.add_argument(false_arg_arg_name, **false_arg_kwargs)
        else:
            raise Exception("Did not know how to handle key " + key +
                            " and value " + str(value))
        self.arg_parser.add_argument(arg_name, **kwargs)

    def configure_arg_parser(self) -> None:
        schema = self.config_json_schema
        if schema['type'] == 'object':
            props = schema['properties']
            if not isinstance(props, dict):
                raise TypeError(f"Did not understand {props} in {schema}")
            required_keys_raw = schema.get('required', [])
            if not isinstance(required_keys_raw, list):
                raise TypeError(f"Did not understand {required_keys_raw} in {schema}")
            required_keys: List = required_keys_raw
            self.configure_from_properties(props, required_keys)
        else:
            raise Exception("Did not know how to parse " +
                            str(self.config_json_schema))

    def parse(self, args: Sequence[str]) -> JobConfig:
        if 'JOB_CONFIG_JSON' in os.environ:
            # https://stackoverflow.com/questions/20199126/reading-json-from-a-file
            with open(os.environ['JOB_CONFIG_JSON']) as json_data:
                return json.load(json_data)
        else:
            if self.config_json_schema is None:
                return {}

            self.configure_arg_parser()
            raw_config = vars(self.arg_parser.parse_args(args))
            return arguments_output_to_config(raw_config)



from ..url.resolver import UrlResolver
from ..records.records_format import BaseRecordsFormat
from ..records.load_plan import RecordsLoadPlan
from ..records.records_directory import RecordsDirectory
from ..url.filesystem import FilesystemDirectoryUrl
from contextlib import contextmanager
from ..url.base import BaseDirectoryUrl
from tempfile import TemporaryDirectory
from abc import ABCMeta, abstractmethod
from typing import Optional, Type, Iterator, IO, List
import sqlalchemy
import logging


logger = logging.getLogger(__name__)


class LoaderFromRecordsDirectory(metaclass=ABCMeta):
    def load_failure_exception(self) -> Type[Exception]:
        return sqlalchemy.exc.InternalError

    def best_scheme_to_load_from(self) -> str:
        return 'file'

    @abstractmethod
    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> Optional[int]:
        """Loads the data from the data specified to the RecordsDirectory
        instance named 'directory'.

        Returns number of rows loaded (if database provides that
        info)."""
        ...

    @abstractmethod
    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        """Return true if the specified format is compatible with the load()
        method"""
        ...

    def temporary_loadable_directory_scheme(self) -> str:
        """If we need to provide a temporary location that this database can
        load from with the temporary_loadable_directory_loc() method,, what
        URL scheme will be used?"""
        return 'file'

    @contextmanager
    def temporary_loadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        """Provide a temporary directory which can be used for bulk import to
        this database and clean it up when done"""
        with TemporaryDirectory(prefix='temporary_loadable_directory_loc') as dirname:
            yield FilesystemDirectoryUrl(dirname)

    def has_temporary_loadable_directory_loc(self) -> bool:
        """Returns True if a temporary directory can be provided by
        temporary_loadable_directory_loc()"""
        # The default implementation uses the local filesystem where
        # Records Mover runs, and we assume we can make temporary
        # files.
        return True

    @abstractmethod
    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        """Candidates to look through when negotiating a common records format
        with a records source.  Will be looked through in order, so
        the better formats (higher fidelity and/or more efficient to
        process) should appear first in the list.
        """
        ...


class LoaderFromFileobj(LoaderFromRecordsDirectory, metaclass=ABCMeta):
    url_resolver: UrlResolver

    @abstractmethod
    def load_from_fileobj(self, schema: str, table: str,
                          load_plan: RecordsLoadPlan, fileobj: IO[bytes]) -> Optional[int]:
        """Loads the data from the file stream provided and append to the existing table."""
        ...

    def load_from_records_directory_via_fileobj(self,
                                                schema: str,
                                                table: str,
                                                load_plan: RecordsLoadPlan,
                                                directory: RecordsDirectory) -> Optional[int]:
        all_urls = directory.manifest_entry_urls()

        total_rows = None
        for url in all_urls:
            loc = self.url_resolver.file_url(url)
            with loc.open() as f:
                logger.info(f"Loading {url} into {schema}.{table}...")
                out = self.load_from_fileobj(schema, table, load_plan, f)
                if out is not None:
                    if total_rows is None:
                        total_rows = 0
                    total_rows += out
        return total_rows

    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> Optional[int]:
        # Implement load() in terms of load_from_fileobj() by default
        return self.load_from_records_directory_via_fileobj(schema=schema,
                                                            table=table,
                                                            load_plan=load_plan,
                                                            directory=directory)



from urllib.request import urlopen
from typing import IO


class UrllibFileMixin:
    url: str

    def open(self, mode: str = "rb") -> IO[bytes]:
        if mode != 'rb':
            raise NotImplementedError(f"Mode {mode} not supported on {self.url}")
        return urlopen(self.url)



import sqlalchemy
from contextlib import contextmanager
from typing import List, Iterator, Optional, Union, Tuple
import logging
from google.cloud.bigquery.dbapi.connection import Connection
from google.cloud.bigquery.client import Client
from google.cloud.bigquery.job import ExtractJobConfig
from records_mover.db.unloader import Unloader
from records_mover.records.records_format import BaseRecordsFormat, AvroRecordsFormat
from records_mover.url.base import BaseDirectoryUrl
from records_mover.url.resolver import UrlResolver
from records_mover.records.unload_plan import RecordsUnloadPlan
from records_mover.records.records_directory import RecordsDirectory
from records_mover.db.errors import NoTemporaryBucketConfiguration

logger = logging.getLogger(__name__)


class BigQueryUnloader(Unloader):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine],
                 url_resolver: UrlResolver,
                 gcs_temp_base_loc: Optional[BaseDirectoryUrl])\
            -> None:
        self.db = db
        self.url_resolver = url_resolver
        self.gcs_temp_base_loc = gcs_temp_base_loc
        super().__init__(db=db)

    def can_unload_format(self, target_records_format: BaseRecordsFormat) -> bool:
        if isinstance(target_records_format, AvroRecordsFormat):
            return True
        return False

    def can_unload_to_scheme(self, scheme: str) -> bool:
        if scheme == 'gs':
            return True
        # Otherwise we'll need a temporary bucket configured for
        # BigQuery to unload into
        return self.gcs_temp_base_loc is not None

    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        return [AvroRecordsFormat()]

    @contextmanager
    def temporary_unloadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        if self.gcs_temp_base_loc is None:
            raise NoTemporaryBucketConfiguration('Please provide a scratch GCS URL in your config '
                                                 '(e.g., set SCRATCH_GCS_URL to a gs:// URL)')
        else:
            with self.gcs_temp_base_loc.temporary_directory() as temp_loc:
                yield temp_loc

    def _parse_bigquery_schema_name(self, schema: str) -> Tuple[Optional[str], str]:
        # https://github.com/mxmzdlv/pybigquery/blob/master/pybigquery/sqlalchemy_bigquery.py#L320
        dataset = None
        project = None

        schema_split = schema.split('.')
        if len(schema_split) == 1:
            dataset, = schema_split
        elif len(schema_split) == 2:
            project, dataset = schema_split
        else:
            raise ValueError(f"Could not understand schema name {schema}")

        return (project, dataset)

    def _extract_job_config(self, unload_plan: RecordsUnloadPlan) -> ExtractJobConfig:
        config = ExtractJobConfig()
        if isinstance(unload_plan.records_format, AvroRecordsFormat):
            config.destination_format = 'AVRO'
            # https://cloud.google.com/bigquery/docs/loading-data-cloud-storage-avro#logical_types
            config.use_avro_logical_types = True
        else:
            raise NotImplementedError(f'Please add support for {unload_plan.records_format}')
        return config

    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> Optional[int]:
        if directory.scheme != 'gs':
            with self.temporary_unloadable_directory_loc() as temp_gcs_loc:
                temp_directory = RecordsDirectory(temp_gcs_loc)
                out = self.unload(schema=schema,
                                  table=table,
                                  unload_plan=unload_plan,
                                  directory=temp_directory)
                temp_directory.copy_to(directory.loc)
                return out
        logger.info("Loading from records directory into BigQuery")
        # https://googleapis.github.io/google-cloud-python/latest/bigquery/usage/tables.html#creating-a-table
        connection: Connection =\
            self.db.engine.raw_connection().connection
        # https://google-cloud.readthedocs.io/en/latest/bigquery/generated/google.cloud.bigquery.client.Client.html
        client: Client = connection._client
        project_id, dataset_id = self._parse_bigquery_schema_name(schema)
        job_config = self._extract_job_config(unload_plan)

        records_format = unload_plan.records_format
        filename = records_format.generate_filename('output')
        destination_uri = directory.loc.file_in_this_directory(filename)
        job = client.extract_table(f"{schema}.{table}",
                                   destination_uri.url,
                                   # Must match the destination dataset location.
                                   job_config=job_config)
        job.result()  # Waits for table load to complete.
        logger.info(f"Unloaded from {dataset_id}:{table} into {filename}")
        directory.save_preliminary_manifest()
        return None



from .hint import LiteralHint, StringHint, Hint
from .types import (
    HintHeaderRow, HintCompression, HintQuoting,
    HintDoublequote, HintEscape, HintEncoding,
    UntypedRecordsHints, PartialRecordsHints
)
from enum import Enum
import logging


logger = logging.getLogger(__name__)


class Hints(Enum):
    value: Hint
    datetimeformattz = StringHint("datetimeformattz",
                                  "YYYY-MM-DD HH24:MI:SSOF",
                                  description=("Format used to write "
                                               "'datetimetz' values"))
    datetimeformat = StringHint("datetimeformat",
                                default="YYYY-MM-DD HH24:MI:SS",
                                description=("Format used to write "
                                             "'datetime' values"))
    # mypy gives this when we pass the HintBlahBlah aliases in as an
    # argument here:
    #
    # error: The type alias to Union is invalid in runtime context
    #
    # Nonetheless, the validation works.
    compression = LiteralHint[HintCompression](HintCompression,  # type: ignore
                                               'compression',
                                               default=None,
                                               description='Compression type of the file.')
    quoting = LiteralHint[HintQuoting](HintQuoting,  # type: ignore
                                       'quoting',
                                       default='minimal',
                                       description=('How quotes are applied to individual fields. '
                                                    'all: quote all fields. '
                                                    'minimal: quote only fields that contain '
                                                    'ambiguous characters (the '
                                                    'delimiter, the escape character, or a line '
                                                    'terminator). '
                                                    'default: never quote fields.'))
    escape = LiteralHint[HintEscape](HintEscape,  # type: ignore
                                     'escape',
                                     default='\\',
                                     description="Character used to escape strings")
    encoding = LiteralHint[HintEncoding](HintEncoding,  # type: ignore
                                         'encoding',
                                         default='UTF8',
                                         description="Text encoding of file")
    dateformat = StringHint('dateformat',
                            default='YYYY-MM-DD',
                            description=("Format used to write "
                                         "'date' values"))
    timeonlyformat = StringHint('timeonlyformat',
                                default="HH24:MI:SS",
                                description=("Format used to write "
                                             "'time' values"))
    # https://docs.python.org/3/library/csv.html#csv.Dialect.doublequote
    doublequote = LiteralHint[HintDoublequote](HintDoublequote,  # type: ignore
                                               'doublequote',
                                               default=False,
                                               description=('Controls how instances of quotechar '
                                                            'appearing inside a field should '
                                                            'themselves be quoted. When True, the '
                                                            'character is doubled. When False, the '
                                                            'escapechar is used as a prefix to the '
                                                            'quotechar.'))
    header_row = LiteralHint[HintHeaderRow](HintHeaderRow,  # type: ignore
                                            'header-row',
                                            default=True,
                                            description=('True if a header row is provided in '
                                                         'the delimited files.'))
    # https://docs.python.org/3/library/csv.html#csv.Dialect.quotechar
    quotechar = StringHint('quotechar',
                           default='"',
                           description=('A one-character string used to quote fields containing '
                                        'special characters, such as the delimiter or quotechar, '
                                        'or which contain new-line characters.'))
    record_terminator = StringHint('record-terminator',
                                   default='\n',
                                   description='String used to close out individual rows of data.')
    field_delimiter = StringHint('field-delimiter',
                                 default=',',
                                 description='Character used between fields.')


def validate_partial_hints(untyped_hints: UntypedRecordsHints,
                           fail_if_cant_handle_hint: bool) -> PartialRecordsHints:
    typed_records_hints: PartialRecordsHints = {}
    for key in untyped_hints:
        hint_obj = Hints[key.replace('-', '_')].value
        value = hint_obj.validate(untyped_hints,
                                  fail_if_cant_handle_hint=fail_if_cant_handle_hint)
        typed_records_hints[key] = value  # type: ignore
    return typed_records_hints



from contextlib import contextmanager
import os
from urllib.parse import urlparse, unquote
import tempfile
from pathlib import Path
from .base import BaseDirectoryUrl, BaseFileUrl
from typing import IO, Iterator, List, Union, Optional


class BaseFilesystemUrl:
    def __init__(self, url: str, **kwargs) -> None:
        self._parsed_url = urlparse(url)
        self.scheme = self._parsed_url.scheme
        if self._parsed_url.netloc != '':
            raise ValueError(f"Could not understand URL {url}")
        # https://docs.python.org/2/library/urlparse.html
        #
        # "The components are not broken up in smaller parts (for
        # example, the network location is a single string), and %
        # escapes are not expanded."
        self.local_file_path = unquote(self._parsed_url.path)
        self.url = url


class FilesystemFileUrl(BaseFilesystemUrl, BaseFileUrl):
    def __init__(self, url: str, **kwargs) -> None:
        super().__init__(url=url, **kwargs)
        self.basename = os.path.basename(self.local_file_path)

    def open(self, mode: str = "rb") -> IO[bytes]:
        return open(self.local_file_path, mode)

    def rename_to(self, new: 'BaseFileUrl') -> 'FilesystemFileUrl':
        if not isinstance(new, FilesystemFileUrl):
            raise TypeError(f'Can only rename to same type, not {new}')
        os.rename(self.local_file_path, new.local_file_path)
        return new

    def filename(self) -> str:
        "Filename, stripped of any directory information"
        return self.basename

    def delete(self) -> None:
        os.remove(self.local_file_path)

    def size(self) -> int:
        statinfo = os.stat(self.local_file_path)
        return statinfo.st_size

    def containing_directory(self) -> 'FilesystemDirectoryUrl':
        parent_dir = os.path.dirname(self.local_file_path)
        new_url = Path(parent_dir).as_uri()
        return FilesystemDirectoryUrl(new_url)


class FilesystemDirectoryUrl(BaseFilesystemUrl, BaseDirectoryUrl):
    def is_directory(self) -> bool:
        return True

    def file_in_this_directory(self, filename: str) -> 'FilesystemFileUrl':
        new_local_file_path = os.path.join(self.local_file_path, filename)
        new_url = Path(new_local_file_path).as_uri()
        return FilesystemFileUrl(new_url)

    def directory_in_this_directory(self, filename: str) -> 'FilesystemDirectoryUrl':
        new_local_file_path = os.path.join(self.local_file_path, filename)
        new_url = Path(new_local_file_path).as_uri()
        return FilesystemDirectoryUrl(new_url)

    def files_in_directory(self) -> List['BaseFileUrl']:
        return [
            self.file_in_this_directory(f)
            for f in os.listdir(self.local_file_path)
            if os.path.isfile(f"{self.local_file_path}/{f}")
        ]

    def files_and_directories_in_directory(self) -> List[Union[BaseFileUrl, BaseDirectoryUrl]]:
        def entry(filename: str) -> Optional[Union[BaseFileUrl, BaseDirectoryUrl]]:
            path = os.path.join(self.local_file_path, filename)
            if os.path.isfile(path):
                return self.file_in_this_directory(filename)
            elif os.path.isdir(path):
                return self.directory_in_this_directory(filename)
            else:
                return None

        return [
            f
            for f in [entry(filename) for filename in os.listdir(self.local_file_path)]
            if f is not None
        ]

    @contextmanager
    def temporary_directory(self: 'FilesystemDirectoryUrl') -> Iterator['FilesystemDirectoryUrl']:
        with tempfile.TemporaryDirectory(dir=self.local_file_path,
                                         prefix="filesystem_py_") as tempdir:
            new_url = Path(tempdir).as_uri()
            yield FilesystemDirectoryUrl(new_url)



import csv
from ...utils import quiet_remove
from ..delimited import cant_handle_hint
from ..processing_instructions import ProcessingInstructions
from ..records_format import DelimitedRecordsFormat
from records_mover.mover_types import _assert_never
import logging
from typing import Set, Dict


logger = logging.getLogger(__name__)


def pandas_to_csv_options(records_format: DelimitedRecordsFormat,
                          unhandled_hints: Set[str],
                          processing_instructions: ProcessingInstructions) -> Dict[str, object]:
    # https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_csv.html
    hints = records_format.\
        validate(fail_if_cant_handle_hint=processing_instructions.fail_if_cant_handle_hint)

    fail_if_cant_handle_hint = processing_instructions.fail_if_cant_handle_hint

    pandas_options: Dict[str, object] = {}

    pandas_options['encoding'] = hints.encoding
    quiet_remove(unhandled_hints, 'encoding')

    if hints.compression is None:
        # hints['compression']=None will output an uncompressed csv,
        # which is the pandas default.
        pass
    elif hints.compression == 'GZIP':
        pandas_options['compression'] = 'gzip'
    elif hints.compression == 'BZIP':
        pandas_options['compression'] = 'bz2'
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', hints)
    quiet_remove(unhandled_hints, 'compression')

    if hints.quoting is None:
        pandas_options['quoting'] = csv.QUOTE_NONE
    elif hints.quoting == 'all':
        pandas_options['quoting'] = csv.QUOTE_ALL
    elif hints.quoting == 'minimal':
        pandas_options['quoting'] = csv.QUOTE_MINIMAL
    elif hints.quoting == 'nonnumeric':
        pandas_options['quoting'] = csv.QUOTE_NONNUMERIC
    else:
        _assert_never(hints.quoting)
    quiet_remove(unhandled_hints, 'quoting')

    pandas_options['doublequote'] = hints.doublequote
    quiet_remove(unhandled_hints, 'doublequote')
    pandas_options['quotechar'] = hints.quotechar
    quiet_remove(unhandled_hints, 'quotechar')

    if hints.escape is None:
        pass
    else:
        pandas_options['escapechar'] = hints.escape
    quiet_remove(unhandled_hints, 'escape')

    pandas_options['header'] = hints.header_row
    quiet_remove(unhandled_hints, 'header-row')

    # Note the limitation on Pandas export with BigQuery around
    # datetimeformattz:
    #
    # https://github.com/bluelabsio/records-mover/issues/95

    # Pandas only gives us one parameter to set for formatting of its
    # Timestamp values, so we need the datetimeformat and
    # datetimeformattz hints to be nearly identical, modulo the
    # timezone at the end which will only appear if it is set in the
    # source data anyway:

    canonical_datetimeformattz = hints.datetimeformattz.replace('HH24', 'HH')
    canonical_datetimeformat = hints.datetimeformat.replace('HH24', 'HH')
    equivalent_with_timezone = f"{canonical_datetimeformat}OF"

    if (canonical_datetimeformattz not in
       [canonical_datetimeformat, equivalent_with_timezone]):
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformat', hints)

    if 'AM' in hints.datetimeformattz:
        hour_specifier = '%I'
    else:
        hour_specifier = '%H'

    pandas_options['date_format'] = hints.datetimeformattz\
        .replace('YYYY', '%Y')\
        .replace('YY', '%y')\
        .replace('MM', '%m')\
        .replace('DD', '%d')\
        .replace('HH24', '%H')\
        .replace('HH12', '%I')\
        .replace('HH', hour_specifier)\
        .replace('MI', '%M')\
        .replace('SS', '%S.%f')\
        .replace('OF', '%z')\
        .replace('AM', '%p')
    quiet_remove(unhandled_hints, 'datetimeformat')
    quiet_remove(unhandled_hints, 'datetimeformattz')

    # timeonlyformat and dateformat are handled in prep_for_csv.py and
    # raw times and dates never appear in dataframes passed a
    # .to_csv() call.
    quiet_remove(unhandled_hints, 'timeonlyformat')
    quiet_remove(unhandled_hints, 'dateformat')

    pandas_options['sep'] = hints.field_delimiter
    quiet_remove(unhandled_hints, 'field-delimiter')

    pandas_options['line_terminator'] = hints.record_terminator
    quiet_remove(unhandled_hints, 'record-terminator')

    return pandas_options



from .base import SupportsRecordsDirectory, SupportsToFileobjsSource
from .fileobjs import FileobjsSource
from contextlib import ExitStack
from ..records_directory import RecordsDirectory
from ..records_format import DelimitedRecordsFormat
from .. import PartialRecordsHints
from contextlib import contextmanager
from typing import Iterator, Optional
from ..processing_instructions import ProcessingInstructions
from ..records_format import BaseRecordsFormat
from ...url.resolver import UrlResolver


class RecordsDirectoryRecordsSource(SupportsRecordsDirectory,
                                    SupportsToFileobjsSource):
    def __init__(self,
                 directory: RecordsDirectory,
                 fail_if_dont_understand: bool,
                 url_resolver: UrlResolver,
                 override_hints: PartialRecordsHints = {}) -> None:
        self.records_format = directory.load_format(fail_if_dont_understand=fail_if_dont_understand)
        if isinstance(self.records_format, DelimitedRecordsFormat):
            self.records_format = self.records_format.alter_hints(override_hints)
        self.directory = directory
        self.url_resolver = url_resolver

    def records_directory(self) -> RecordsDirectory:
        return self.directory

    @contextmanager
    def to_fileobjs_source(self,
                           processing_instructions: ProcessingInstructions,
                           records_format_if_possible: Optional[BaseRecordsFormat] = None)\
            -> Iterator[FileobjsSource]:

        all_urls = self.directory.manifest_entry_urls()

        locs = [self.url_resolver.file_url(url) for url in all_urls]

        with ExitStack() as stack:
            target_names_to_input_fileobjs = {
                loc.filename(): stack.enter_context(loc.open())
                for loc in locs
            }
            records_schema = self.directory.load_schema_json_obj()
            with FileobjsSource.infer_if_needed(target_names_to_input_fileobjs,
                                                processing_instructions=processing_instructions,
                                                records_format=self.records_format,
                                                records_schema=records_schema,
                                                initial_hints=None) as f:
                yield f



import sqlalchemy as sa
from records_mover.db import create_sqlalchemy_url
from airflow.hooks.base import BaseHook


class SqlAlchemyDbHook(BaseHook):
    "Airflow hook to provide a SQLAlchemy engine from an Airflow database connection ID"

    def __init__(self, db_conn_id: str) -> None:
        """
        :param db_conn_id: Airflow connection ID for the database which should be connected to.
        """
        self.db_conn_id = db_conn_id

    def get_conn(self) -> sa.engine.Engine:
        """
        :return: SQLAlchemy engine corresponding to this Airflow database connection ID.
        """
        conn = BaseHook.get_connection(self.db_conn_id)
        db_url = create_sqlalchemy_url(
            {
                'host': conn.host,
                'port': conn.port,
                'database': conn.schema,
                'user': conn.login,
                'password': conn.password,
                'type': conn.extra_dejson.get('type', conn.conn_type.lower()),
            }
        )
        return sa.create_engine(db_url)



from contextlib import contextmanager
from sqlalchemy.schema import Table
from .sqlalchemy_postgres_copy import copy_to
from ..quoting import quote_value
from ...records.unload_plan import RecordsUnloadPlan
from ...records.records_format import BaseRecordsFormat
from ...records.records_directory import RecordsDirectory
from ...records.records_format import DelimitedRecordsFormat
from ...records.delimited import complain_on_unhandled_hints
from records_mover.url.base import BaseDirectoryUrl
from records_mover.url.filesystem import FilesystemDirectoryUrl
from typing import List, Iterator
from tempfile import TemporaryDirectory
from ..unloader import Unloader
from .copy_options import postgres_copy_to_options
import logging

logger = logging.getLogger(__name__)


class PostgresUnloader(Unloader):
    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> None:
        if not isinstance(unload_plan.records_format, DelimitedRecordsFormat):
            raise NotImplementedError("This only supports delimited mode for now")

        unhandled_hints = set(unload_plan.records_format.hints.keys())
        processing_instructions = unload_plan.processing_instructions
        date_output_style, date_order_style, postgres_options =\
            postgres_copy_to_options(unhandled_hints,
                                     unload_plan.records_format,
                                     processing_instructions.fail_if_cant_handle_hint)
        if date_order_style is None:
            # U-S-A!  U-S-A!
            date_order_style = 'MDY'
        complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                    unhandled_hints,
                                    unload_plan.records_format.hints)

        table_obj = Table(table,
                          self.meta,
                          schema=schema,
                          autoload=True,
                          autoload_with=self.db)

        with self.db.engine.begin() as conn:
            # https://www.postgresql.org/docs/8.3/sql-set.html
            #
            # The effects of SET LOCAL last only till the end of the
            # current transaction, whether committed or not. A special
            # case is SET followed by SET LOCAL within a single
            # transaction: the SET LOCAL value will be seen until the end
            # of the transaction, but afterwards (if the transaction is
            # committed) the SET value will take effect.
            date_style = f"{date_output_style}, {date_order_style}"
            sql = f"SET LOCAL DateStyle = {quote_value(conn, date_style)}"
            logger.info(sql)
            conn.execute(sql)

            filename = unload_plan.records_format.generate_filename('data')
            loc = directory.loc.file_in_this_directory(filename)
            with loc.open(mode='wb') as fileobj:
                copy_to(table_obj.select(),
                        fileobj,
                        conn,
                        **postgres_options)

        logger.info('Copy complete')
        directory.save_preliminary_manifest()

    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        return [
            #
            # Notes:
            #
            # The 'vertica' variant won't work, as the
            # record-delimiter in Postgres unloads can't be anything
            # other than a UNIX newline.
            #
            # The 'bigquery' variant isn't feasible, as there's no way
            # to export the datetimetz equivalent type without a
            # timezone indicator.
            #
            # Postgres doesn't support compression on unload, so we
            # need to avoid that.
            #
            DelimitedRecordsFormat(variant='bluelabs',
                                   hints={'compression': None}),
            #
            # Postgres supports a limited number of export date
            # styles, so only a subvariant of the csv variant is
            # feasible.  See copy_options/date_output_style.py for
            # details.
            #
            DelimitedRecordsFormat(variant='csv',
                                   hints={
                                       'compression': None,
                                       'dateformat': 'YYYY-MM-DD',
                                       'timeonlyformat': 'HH24:MI:SS',
                                       'datetimeformattz': 'YYYY-MM-DD HH24:MI:SSOF',
                                       'datetimeformat': 'YYYY-MM-DD HH24:MI:SS',
                                   }),
        ]

    def can_unload_to_scheme(self, scheme: str) -> bool:
        # Unloading is done via streams, so it is scheme-independent
        # and requires no scratch buckets.
        return True

    @contextmanager
    def temporary_unloadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        with TemporaryDirectory(prefix='temporary_unloadable_directory_loc') as dirname:
            yield FilesystemDirectoryUrl(dirname)

    def can_unload_format(self, target_records_format: BaseRecordsFormat) -> bool:
        try:
            unload_plan = RecordsUnloadPlan(records_format=target_records_format)
            records_format = unload_plan.records_format
            if not isinstance(records_format, DelimitedRecordsFormat):
                return False
            unhandled_hints = set(records_format.hints.keys())
            processing_instructions = unload_plan.processing_instructions
            postgres_copy_to_options(unhandled_hints,
                                     records_format,
                                     processing_instructions.fail_if_cant_handle_hint)
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints,
                                        records_format.hints)
            return True
        except NotImplementedError:
            return False



from pandas import DataFrame
import datetime
import pandas as pd
from records_mover.records import ProcessingInstructions
from records_mover.records.schema import RecordsSchema
from records_mover.records.schema.field import RecordsSchemaField
from records_mover.records import DelimitedRecordsFormat
from records_mover.records.delimited import (python_date_format_from_hints,
                                             python_time_format_from_hints,
                                             cant_handle_hint)
import logging
from typing import Optional, Union, TypeVar

logger = logging.getLogger(__name__)


T = TypeVar('T', bound=Union[pd.Series, pd.Index])


def _convert_series_or_index(series_or_index: T,
                             field: RecordsSchemaField,
                             records_format: DelimitedRecordsFormat,
                             processing_instructions: ProcessingInstructions) -> Optional[T]:
    if field.field_type == 'date':
        if (isinstance(series_or_index[0], pd.Timestamp) or
           isinstance(series_or_index[0], datetime.date)):
            logger.info(f"Converting {series_or_index.name} from np.datetime64 to "
                        "string in CSV's format")
            logger.debug("Dtype is %s, first element type %s", series_or_index.dtype,
                         type(series_or_index[0]))
            hint_date_format = records_format.hints['dateformat']
            assert isinstance(hint_date_format, str)
            pandas_date_format = python_date_format_from_hints.get(hint_date_format)
            if pandas_date_format is None:
                cant_handle_hint(processing_instructions.fail_if_cant_handle_hint,
                                 'dateformat',
                                 records_format.hints)
                pandas_date_format = '%Y-%m-%d'
            if isinstance(series_or_index[0], pd.Timestamp):
                if isinstance(series_or_index, pd.Series):
                    return series_or_index.dt.strftime(pandas_date_format)
                else:
                    return series_or_index.strftime(pandas_date_format)
            else:
                return series_or_index.apply(pd.Timestamp).dt.strftime(pandas_date_format)
    elif field.field_type == 'time':
        if (not (isinstance(series_or_index[0], pd.Timestamp) or
                 isinstance(series_or_index[0], datetime.time))):
            logger.warning(f"Found {series_or_index.name} as unexpected "
                           f"type {type(series_or_index[0])}")
        else:
            logger.info(f"Converting {series_or_index.name} from np.datetime64 to string "
                        "in CSV's format")
            logger.debug("Dtype is %s, first element type %s", series_or_index.dtype,
                         type(series_or_index[0]))
            hint_time_format = records_format.hints['timeonlyformat']
            assert isinstance(hint_time_format, str)
            pandas_time_format = python_time_format_from_hints.get(hint_time_format)
            if pandas_time_format is None:
                cant_handle_hint(processing_instructions.fail_if_cant_handle_hint,
                                 'timeonlyformat',
                                 records_format.hints)
                pandas_time_format = '%H:%M:%S'

            if isinstance(series_or_index[0], datetime.time):
                return series_or_index.apply(lambda d: d.strftime(pandas_time_format))
            elif isinstance(series_or_index, pd.Series):
                return series_or_index.dt.strftime(pandas_time_format)
            else:
                return series_or_index.strftime(pandas_time_format)
    else:
        logger.debug(f"Not converting field type {field.field_type}")

    return None


def prep_df_for_csv_output(df: DataFrame,
                           include_index: bool,
                           records_schema: RecordsSchema,
                           records_format: DelimitedRecordsFormat,
                           processing_instructions: ProcessingInstructions) -> DataFrame:
    #
    # Pandas dataframes only have a native 'datetime'/'datetimetz'
    # datatype (pd.Timestamp), not an individal 'date', 'time' or
    # 'timetz' class.  To generate the correct thing when writing out
    # a 'date' or 'time' type to a CSV with Pandas' .to_csv() method,
    # we need to convert those values to strings that represent
    # exactly what we want.
    #
    # An example of when we can get these pd.Timestamp values inside a
    # dataframe from read_csv() is when we tell it that a given column
    # represents a date and/or time, allowing us to pick the format on
    # the way out.
    #
    formatted_df = df.copy(deep=False)
    remaining_fields = records_schema.fields.copy()
    if include_index:
        field = remaining_fields.pop(0)
        formatted_index = _convert_series_or_index(formatted_df.index,
                                                   field,
                                                   records_format,
                                                   processing_instructions)
        if formatted_index is not None:
            formatted_df.index = formatted_index

    for index, field in enumerate(remaining_fields):
        series = formatted_df.iloc[:, index]
        formatted_series = _convert_series_or_index(series,
                                                    field,
                                                    records_format,
                                                    processing_instructions)
        if formatted_series is not None:
            formatted_df.iloc[:, index] = formatted_series
    return formatted_df



import sqlalchemy
from pathlib import Path
from records_mover.records import ProcessingInstructions
from records_mover.db.loader import LoaderFromRecordsDirectory
from records_mover.url.filesystem import FilesystemDirectoryUrl, FilesystemFileUrl
from records_mover.records.load_plan import RecordsLoadPlan
from records_mover.records.records_directory import RecordsDirectory
from records_mover.records.records_format import BaseRecordsFormat, DelimitedRecordsFormat
from .load_options import mysql_load_options
from ...records.delimited import complain_on_unhandled_hints
from ...url.resolver import UrlResolver
from typing import Union, List
import logging
import tempfile

logger = logging.getLogger(__name__)


class MySQLLoader(LoaderFromRecordsDirectory):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
                 url_resolver: UrlResolver) -> None:
        self.db = db
        self.url_resolver = url_resolver

    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> None:
        if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
            raise NotImplementedError('Teach me how to load '
                                      f'{load_plan.records_format.format_type} format')

        unhandled_hints = set(load_plan.records_format.hints.keys())
        processing_instructions = load_plan.processing_instructions
        load_options = mysql_load_options(unhandled_hints,
                                          load_plan.records_format,
                                          processing_instructions.fail_if_cant_handle_hint)
        complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                    unhandled_hints, load_plan.records_format.hints)
        if not isinstance(directory.loc, FilesystemDirectoryUrl):
            with tempfile.TemporaryDirectory(prefix='mysql_loader_load') as tempdir:
                filesystem_url = Path(tempdir).as_uri()
                temp_filesystem_loc = FilesystemDirectoryUrl(filesystem_url)
                filesystem_directory = directory.copy_to(temp_filesystem_loc)
                return self.load(schema=schema,
                                 table=table,
                                 load_plan=load_plan,
                                 directory=filesystem_directory)

        all_urls = directory.manifest_entry_urls()

        locs = [self.url_resolver.file_url(url) for url in all_urls]
        for loc in locs:
            # This came from a FilesystemDirectoryUrl, so it had better be...
            assert isinstance(loc, FilesystemFileUrl)
            filename = loc.local_file_path
            sql = load_options.generate_load_data_sql(filename=filename,
                                                      table_name=table,
                                                      schema_name=schema)
            logger.info(f"Loading to MySQL with options: {load_options}")
            logger.info(str(sql))
            self.db.execute(sql)
            logger.info("MySQL LOAD DATA complete.")
        return None

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        try:
            processing_instructions = ProcessingInstructions()
            load_plan = RecordsLoadPlan(records_format=source_records_format,
                                        processing_instructions=processing_instructions)
            if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
                return False

            unhandled_hints = set(load_plan.records_format.hints.keys())
            processing_instructions = load_plan.processing_instructions
            mysql_load_options(unhandled_hints, load_plan.records_format,
                               fail_if_cant_handle_hint=True)
            complain_on_unhandled_hints(fail_if_dont_understand=True,
                                        unhandled_hints=unhandled_hints,
                                        hints=load_plan.records_format.hints)
            return True
        except NotImplementedError:
            return False

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        return [
            # MySQL supports a healthy amount of load types, but
            # doesn't support loading compressed files.
            DelimitedRecordsFormat(variant='bluelabs',
                                   hints={
                                       'compression': None
                                   }),
            DelimitedRecordsFormat(variant='bigquery',
                                   hints={
                                       'compression': None
                                   }),
            DelimitedRecordsFormat(variant='vertica',
                                   hints={
                                       'compression': None
                                   }),
        ]



from .json_schema_document import JsonSchemaDocument


class JsonParameter:
    def __init__(self,
                 name: str,
                 json_schema_document: JsonSchemaDocument,
                 optional: bool = False) -> None:
        self.name = name
        self.optional = optional
        self.json_schema_document = json_schema_document



from records_mover.db import DBDriver
from records_mover.records.prep import TablePrep, TargetTableDetails
from records_mover.records.processing_instructions import ProcessingInstructions
from records_mover.records.results import MoveResult
from records_mover.records.prep_and_load import prep_and_load
from records_mover.records.load_plan import RecordsLoadPlan
from records_mover.records.sources.fileobjs import FileobjsSource
from records_mover.records.targets.table.base import BaseTableMoveAlgorithm
from records_mover.utils.concat_files import ConcatFiles
from typing import Optional, IO
import logging

logger = logging.getLogger(__name__)


class DoMoveFromFileobjsSource(BaseTableMoveAlgorithm):
    fileobj: IO[bytes]

    def __init__(self,
                 prep: TablePrep,
                 target_table_details: TargetTableDetails,
                 fileobjs_source: FileobjsSource,
                 processing_instructions: ProcessingInstructions) -> None:
        self.fileobjs_source = fileobjs_source
        all_fileobjs = list(self.fileobjs_source.target_names_to_input_fileobjs.values())
        if len(all_fileobjs) != 1:
            self.fileobj = ConcatFiles(all_fileobjs)  # type: ignore
        else:
            self.fileobj = all_fileobjs[0]
        self.records_format = self.fileobjs_source.records_format
        self.plan = RecordsLoadPlan(records_format=self.records_format,
                                    processing_instructions=processing_instructions)
        super().__init__(prep, target_table_details, processing_instructions)

    def load(self, driver: DBDriver) -> Optional[int]:
        loader_from_fileobj = driver.loader_from_fileobj()
        # This is only reached in move() when
        # records_target.can_move_from_fileobjs_source() is true,
        # which is only true when .load_from_fileobj() is not None.
        assert loader_from_fileobj is not None
        return loader_from_fileobj.load_from_fileobj(schema=self.tbl.schema_name,
                                                     table=self.tbl.table_name,
                                                     load_plan=self.plan,
                                                     fileobj=self.fileobj)

    def reset_before_reload(self) -> None:
        if not self.tbl.drop_and_recreate_on_load_error:
            raise
        if not self.fileobj.seekable():
            raise NotImplementedError("This process cannot be restarted with this "
                                      "type of file object. Please fix your table "
                                      "manually and try the load again.")
        self.fileobj.seek(0)

    def move(self) -> MoveResult:
        with self.tbl.db_engine.begin() as db:
            driver = self.tbl.db_driver(db)
            schema_obj = self.fileobjs_source.records_schema
            schema_sql = self.schema_sql_for_load(schema_obj, self.records_format, driver)
            loader_from_fileobj = driver.loader_from_fileobj()
            # This is only reached in move() when
            # records_target.can_move_from_fileobjs_source() is true,
            # which is only true when .load_from_fileobj() is not None.
            assert loader_from_fileobj is not None
            load_exception = loader_from_fileobj.load_failure_exception()

        return prep_and_load(self.tbl, self.prep, schema_sql, self.load,
                             load_exception,
                             self.reset_before_reload)



from ...utils import quiet_remove
from ...records.delimited import cant_handle_hint
from typing import Dict, Set
from ...records.load_plan import RecordsLoadPlan
from ...records.records_format import DelimitedRecordsFormat


VerticaImportOptions = Dict[str, object]


def vertica_import_options(unhandled_hints: Set[str],
                           load_plan: RecordsLoadPlan) -> VerticaImportOptions:
    fail_if_cant_handle_hint = load_plan.processing_instructions.fail_if_cant_handle_hint
    if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
        raise NotImplementedError("Not currently able to import "
                                  f"{load_plan.records_format.format_type}")
    hints = load_plan.records_format.hints

    vertica_options: VerticaImportOptions = {}

    vertica_options['record_terminator'] = hints['record-terminator']
    quiet_remove(unhandled_hints, 'record-terminator')
    if hints['compression'] is None:
        vertica_options['gzip'] = False
    elif hints['compression'] == 'GZIP':
        vertica_options['gzip'] = True
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', hints)

    quiet_remove(unhandled_hints, 'compression')
    vertica_options['escape_as'] = hints['escape']
    quiet_remove(unhandled_hints, 'escape')

    if not hints['timeonlyformat'] in ['HH24:MI:SS',
                                       'HH:MI:SS',
                                       'HH12:MI AM']:
        # Vertica seems to be able to understand these on import automatically
        cant_handle_hint(fail_if_cant_handle_hint, 'timeonlyformat', hints)
    quiet_remove(unhandled_hints, 'timeonlyformat')

    if hints['quoting'] is None:
        vertica_options['enclosed_by'] = None
    elif hints['quoting'] == 'all':
        if hints['doublequote'] is not False:
            cant_handle_hint(fail_if_cant_handle_hint, 'doublequote', hints)
        vertica_options['enclosed_by'] = hints['quotechar']
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'quoting', hints)
    quiet_remove(unhandled_hints, 'doublequote')
    quiet_remove(unhandled_hints, 'quotechar')
    quiet_remove(unhandled_hints, 'quoting')

    if not hints['dateformat'] in ['YYYY-MM-DD']:
        cant_handle_hint(fail_if_cant_handle_hint, 'dateformat', hints)
    quiet_remove(unhandled_hints, 'dateformat')

    vertica_options['delimiter'] = hints['field-delimiter']
    quiet_remove(unhandled_hints, 'field-delimiter')

    if hints['datetimeformattz'] not in ['YYYY-MM-DD HH:MI:SSOF',
                                         'YYYY-MM-DD HH24:MI:SSOF']:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformattz', hints)
    quiet_remove(unhandled_hints, 'datetimeformattz')

    if not hints['datetimeformat'] in ['YYYY-MM-DD HH:MI:SS', 'YYYY-MM-DD HH24:MI:SS']:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformat', hints)
    quiet_remove(unhandled_hints, 'datetimeformat')

    if hints['encoding'] != 'UTF8':
        cant_handle_hint(fail_if_cant_handle_hint, 'encoding', hints)
    quiet_remove(unhandled_hints, 'encoding')

    if hints['header-row']:
        vertica_options['skip'] = 1
    else:
        vertica_options['skip'] = 0
    quiet_remove(unhandled_hints, 'header-row')

    vertica_options['load_method'] = 'AUTO'

    vertica_options['no_commit'] = False

    if load_plan.processing_instructions.max_failure_rows is not None:
        vertica_options['trailing_nullcols'] = True
        vertica_options['rejectmax'] = load_plan.processing_instructions.max_failure_rows
        vertica_options['enforcelength'] = None
        vertica_options['error_tolerance'] = None
        vertica_options['abort_on_error'] = None
    elif not load_plan.processing_instructions.fail_if_row_invalid:
        vertica_options['trailing_nullcols'] = True
        vertica_options['rejectmax'] = None
        vertica_options['enforcelength'] = False
        vertica_options['error_tolerance'] = True
        vertica_options['abort_on_error'] = False
    else:
        vertica_options['trailing_nullcols'] = False
        vertica_options['rejectmax'] = 1
        vertica_options['enforcelength'] = True
        vertica_options['error_tolerance'] = False
        vertica_options['abort_on_error'] = True

    vertica_options['null_as'] = None

    return vertica_options



import os
import sys
import logging
import io
from typing import IO, Set, Optional


def _adjusted_log_level(default_log_level: int, name: str) -> int:
    log_level_str = os.environ.get(name.upper() + '_LOG_LEVEL', None)
    if log_level_str is None:
        return default_log_level
    return getattr(logging, log_level_str.upper())


_secrets: Set[str] = set()


def register_secret(secret: Optional[object]) -> None:
    if secret is not None:
        _secrets.add(str(secret))


# https://stackoverflow.com/questions/45469808/how-to-wrap-a-python-text-stream-to-replace-strings-on-the-fly
class SecretsRedactingLogStream(io.TextIOBase):
    def __init__(self, underlying: IO[str]) -> None:
        self.underlying = underlying
        self.secrets = ['records']

    def write(self, s: str) -> int:
        for secret in _secrets:
            if secret in s:
                replacement = '*' * len(secret)
                s = s.replace(secret, replacement)
        return self.underlying.write(s)


def set_stream_logging(name: str = 'records_mover',
                       level: int = logging.INFO,
                       stream: IO[str] = sys.stdout,
                       fmt: str = '%(asctime)s - %(message)s',
                       datefmt: str = '%H:%M:%S') -> None:
    """
    records-mover logs details about its operations using Python logging.  This method is a
    simple way to configure that logging to be output to a stream (by default, stdout).

    You can use it for other things (e.g., dependencies of
    records-mover) by adjusting the 'name' argument.

    :param name: Name of the package to set logging under.  If set to 'foo', you can set a log
      variable FOO_LOG_LEVEL to the log level threshold you'd like to set (INFO/WARNING/etc) - so
      you can by default set, say, export RECORDS_MOVER_LOG_LEVEL=WARNING to quiet down loging, or
      export RECORDS_MOVER_LOG_LEVEL=DEBUG to increase it.
    :param level: Logging more detailed than this will not be output to the stream.
    :param stream: Stream which logging should be sent (e.g., sys.stdout, sys.stdin, or perhaps a
      file you open)
    :param fmt: Logging format to send to Python'slogging.Formatter() - determines what details
      will be sent.
    :param datefmt: Date format to send to Python'slogging.Formatter() - determines how the current
      date/time will be recorded in the log.
    """
    adjusted_level = _adjusted_log_level(level, name)
    logger = logging.getLogger(name)
    logger.setLevel(adjusted_level)
    wrapper = SecretsRedactingLogStream(stream)
    handler = logging.StreamHandler(stream=wrapper)
    handler.setLevel(adjusted_level)
    formatter = logging.Formatter(fmt, datefmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)



from db_facts import db
import os
import base64
import json
from typing import Iterable, Optional
from .base_creds import BaseCreds
from typing import TYPE_CHECKING
from db_facts.db_facts_types import DBFacts
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa
    import boto3  # noqa


class CredsViaEnv(BaseCreds):
    def _gcp_creds_from_env(self,
                            scopes: Iterable[str]) ->\
            Optional['google.auth.credentials.Credentials']:
        if 'GCP_SERVICE_ACCOUNT_JSON_BASE64' not in os.environ:
            return None
        import google.oauth2.service_account
        json_base64 = os.environ['GCP_SERVICE_ACCOUNT_JSON_BASE64']
        json_str = base64.b64decode(json_base64.encode('utf-8'))
        cred_details = json.loads(json_str)

        return google.oauth2.service_account.Credentials.\
            from_service_account_info(cred_details, scopes=scopes)

    def _gcp_creds(self, gcp_creds_name: str,
                   scopes: Iterable[str]) -> 'google.auth.credentials.Credentials':
        # No current way to differentiate between different creds;
        # just on env variable recognized.
        creds_from_env = self._gcp_creds_from_env(scopes)
        if creds_from_env is None:
            raise KeyError('Please set GCP_SERVICE_ACCOUNT_JSON_BASE64 to configure Google '
                           'Cloud Platform credentials')
        return creds_from_env

    def db_facts(self, db_creds_name: str) -> DBFacts:
        return db(db_creds_name.split('-'))

    def _gcp_creds_of_last_resort(self,
                                  scopes: Iterable[str]) ->\
            Optional['google.auth.credentials.Credentials']:
        creds = self._gcp_creds_from_env(scopes)
        if creds is None:
            # Fall back and use Google's default configuration files
            creds = super()._gcp_creds_of_last_resort(scopes=scopes)
        return creds



from typing import Optional
from records_mover.records.prep import TablePrep, TargetTableDetails
from records_mover.records.records_format import BaseRecordsFormat
from records_mover.db import DBDriver
from records_mover.utils.lazyprop import lazyprop
from records_mover.records.records_directory import RecordsDirectory
from records_mover.records.schema import RecordsSchema
from records_mover.records.prep_and_load import prep_and_load
from records_mover.records.processing_instructions import ProcessingInstructions
from records_mover.records.results import MoveResult
from records_mover.records.load_plan import RecordsLoadPlan
from records_mover.records.targets.table.base import BaseTableMoveAlgorithm
import logging

logger = logging.getLogger(__name__)


class DoMoveFromRecordsDirectory(BaseTableMoveAlgorithm):
    def __init__(self,
                 prep: TablePrep,
                 target_table_details: TargetTableDetails,
                 directory: RecordsDirectory,
                 processing_instructions: ProcessingInstructions,
                 override_records_format: Optional[BaseRecordsFormat]) -> None:
        self.directory = directory
        self.override_records_format = override_records_format
        super().__init__(prep, target_table_details, processing_instructions)

    def load_schema_sql(self,
                        driver: 'DBDriver') -> str:
        schema_obj: Optional[RecordsSchema] = self.directory.load_schema_json_obj()
        if schema_obj is None:
            sql = self.directory.load_schema_sql_from_sql_file()
            if sql is None:
                raise SyntaxError(f'RecordsDirectory ({self}) contains '
                                  'neither schema JSON nor schema SQL')
            return sql
        records_format = self.load_plan.records_format
        return self.schema_sql_for_load(schema_obj, records_format, driver)

    @lazyprop
    def load_plan(self) -> RecordsLoadPlan:
        if self.override_records_format is None:
            records_format = self.directory.load_format(self.processing_instructions.
                                                        fail_if_dont_understand)
        else:
            records_format = self.override_records_format
        return RecordsLoadPlan(records_format=records_format,
                               processing_instructions=self.processing_instructions)

    def load(self, driver: DBDriver) -> Optional[int]:
        plan = self.load_plan
        loader = driver.loader()
        # If we've gotten here, .can_move_from_format() has
        # returned True in the move() method, and that can only happen
        # if we have a valid loader.
        assert loader is not None
        return loader.load(schema=self.tbl.schema_name, table=self.tbl.table_name,
                           load_plan=plan, directory=self.directory)

    def move(self) -> MoveResult:
        logger.info("Connecting to database...")

        with self.tbl.db_engine.begin() as db:
            driver = self.tbl.db_driver(db)
            loader = driver.loader()
            # If we've gotten here, .can_move_from_format() has
            # returned True in the move() method, and that can only happen
            # if we have a valid loader.
            assert loader is not None
            load_exception_type = loader.load_failure_exception()
            schema_sql = self.load_schema_sql(driver)
        return prep_and_load(self.tbl, self.prep, schema_sql, self.load,
                             load_exception_type)



import urllib
import vertica_python
import sqlalchemy
from .import_sql import vertica_import_sql
from .records_import_options import vertica_import_options
from .io_base_wrapper import IOBaseWrapper
from ...url.resolver import UrlResolver
from ...records.load_plan import RecordsLoadPlan
from ...records.delimited import complain_on_unhandled_hints
from ...records.records_format import DelimitedRecordsFormat, BaseRecordsFormat
from ...records.processing_instructions import ProcessingInstructions
from ..loader import LoaderFromFileobj
from typing import IO, Union, List, Type
import logging

logger = logging.getLogger(__name__)


class VerticaLoader(LoaderFromFileobj):
    def __init__(self,
                 url_resolver: UrlResolver,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine]) -> None:
        self.url_resolver = url_resolver
        self.db = db

    def load_from_fileobj(self,
                          schema: str,
                          table: str,
                          load_plan: RecordsLoadPlan,
                          fileobj: IO[bytes]) -> None:
        records_format = load_plan.records_format
        if not isinstance(records_format, DelimitedRecordsFormat):
            raise NotImplementedError("Not currently able to load "
                                      f"{records_format.format_type}")
        processing_instructions = load_plan.processing_instructions
        unhandled_hints = set(records_format.hints.keys())
        vertica_options = vertica_import_options(unhandled_hints, load_plan)
        complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                    unhandled_hints,
                                    records_format.hints)

        # vertica_options isn't yet a TypedDict that matches the
        # vertica_import_sql options, so suppress type checking
        import_sql = vertica_import_sql(db_engine=self.db.engine, table=table,
                                        schema=schema, **vertica_options)   # type: ignore
        rawconn = None
        try:
            rawconn = self.db.engine.raw_connection()
            cursor = rawconn.cursor()
            logger.info(import_sql)
            if isinstance(fileobj, urllib.response.addinfourl):
                # Vertica driver is a little too aggressive validating
                # streams and checks type, not behavior, and it
                # rejects some third-party streams.
                #
                # In this case, give it something with the type it's
                # looking for:
                compatible_fileobj = IOBaseWrapper(fileobj)
                cursor.copy(import_sql, compatible_fileobj)
            else:
                cursor.copy(import_sql, fileobj)
            logger.info('Copy complete')
            return None
        finally:
            if rawconn is not None:
                rawconn.close()

    def load_failure_exception(self) -> Type[Exception]:
        return vertica_python.errors.CopyRejected

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        try:
            processing_instructions = ProcessingInstructions()
            load_plan = RecordsLoadPlan(records_format=source_records_format,
                                        processing_instructions=processing_instructions)
            if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
                return False

            unhandled_hints = set(load_plan.records_format.hints.keys())
            processing_instructions = load_plan.processing_instructions
            vertica_import_options(unhandled_hints, load_plan)
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, load_plan.records_format.hints)
            return True
        except NotImplementedError:
            return False

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        return [
            DelimitedRecordsFormat(variant='bluelabs'),
            DelimitedRecordsFormat(variant='vertica')
        ]



from ..driver import DBDriver
import logging
from ...records import RecordsSchema
from ...records.records_format import BaseRecordsFormat, ParquetRecordsFormat, AvroRecordsFormat
from ...utils.limits import INT64_MAX, INT64_MIN, FLOAT64_SIGNIFICAND_BITS, num_digits
import re
from typing import Union, Optional, Tuple
from ...url.resolver import UrlResolver
import sqlalchemy
from .loader import BigQueryLoader
from .unloader import BigQueryUnloader
from ..loader import LoaderFromFileobj, LoaderFromRecordsDirectory
from ..unloader import Unloader
from ...url.base import BaseDirectoryUrl


logger = logging.getLogger(__name__)


class BigQueryDBDriver(DBDriver):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine],
                 url_resolver: UrlResolver,
                 gcs_temp_base_loc: Optional[BaseDirectoryUrl] = None,
                 **kwargs: object) -> None:
        super().__init__(db)
        self._bigquery_loader =\
            BigQueryLoader(db=self.db,
                           url_resolver=url_resolver,
                           gcs_temp_base_loc=gcs_temp_base_loc)
        self._bigquery_unloader =\
            BigQueryUnloader(db=self.db,
                             url_resolver=url_resolver,
                             gcs_temp_base_loc=gcs_temp_base_loc)

    def loader(self) -> Optional[LoaderFromRecordsDirectory]:
        return self._bigquery_loader

    def loader_from_fileobj(self) -> LoaderFromFileobj:
        return self._bigquery_loader

    def unloader(self) -> Unloader:
        return self._bigquery_unloader

    def type_for_date_plus_time(self, has_tz: bool = False) -> sqlalchemy.sql.sqltypes.DateTime:
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types
        if has_tz:
            # pybigquery doesn't set the timezone flag :(
            return sqlalchemy.sql.sqltypes.TIMESTAMP()
        else:
            return sqlalchemy.sql.sqltypes.DATETIME()

    def make_column_name_valid(self, colname: str) -> str:
        # https://cloud.google.com/bigquery/docs/schemas#column_names

        # replace anything which isn't a letter or number with '_',
        # which is the only valid symbol
        return re.sub(r'[^A-Za-z0-9]', '_', colname)

    def integer_limits(self,
                       type_: sqlalchemy.types.Integer) ->\
            Optional[Tuple[int, int]]:
        # Only one int type for BigQuery, represented as sqlalchemy.types.Integer...
        return (INT64_MIN, INT64_MAX)

    def fp_constraints(self,
                       type_: sqlalchemy.types.Float) ->\
            Optional[Tuple[int, int]]:
        # Only one float type for BigQuery, represented as sqlalchemy.types.Float...
        return (64, FLOAT64_SIGNIFICAND_BITS)

    def fixed_point_constraints(self,
                                type_: sqlalchemy.types.Numeric) ->\
            Optional[Tuple[int, int]]:
        from sqlalchemy_bigquery import BIGNUMERIC
        base_type = super().fixed_point_constraints(type_)
        if base_type:
            return base_type
        elif isinstance(type_, sqlalchemy.sql.sqltypes.DECIMAL):
            return (38, 9)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.NUMERIC):
            return (38, 9)
        elif isinstance(type_, BIGNUMERIC):
            return (76, 38)
        else:
            logger.warning(f"Don't know how to handle unexpected BigQuery type {type(type_)}")
            return None

    def type_for_fixed_point(self,
                             precision: int,
                             scale: int) -> sqlalchemy.sql.sqltypes.Numeric:
        # BigQuery now supports precision/scale args and higher precision/scale in the BIGNUMERIC:
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#parameterized_decimal_type
        #
        # sqlalchemy-bigquery uses the Numeric sqlalchemy type and chooses NUMERIC/BIGNUMERIC
        # as needed.
        if scale > 38 or precision > 76:
            logger.warning('Using BigQuery FLOAT64 type to represent '
                           f'NUMERIC({precision},{scale}))')
            return sqlalchemy.types.Float()

        return sqlalchemy.sql.sqltypes.Numeric(precision, scale)

    def type_for_integer(self,
                         min_value: Optional[int],
                         max_value: Optional[int]) -> sqlalchemy.types.TypeEngine:
        """Find correct integral column type to fit the given min and max integer values"""

        if min_value is not None and max_value is not None:
            if min_value >= INT64_MIN and max_value <= INT64_MAX:
                # BigQuery has only 64-bit integers
                return sqlalchemy.sql.sqltypes.Integer()
            else:
                num_digits_min = num_digits(min_value)
                num_digits_max = num_digits(max_value)
                digit_count = max(num_digits_min, num_digits_max)
                return self.type_for_fixed_point(precision=digit_count,
                                                 scale=0)
        return super().type_for_integer(min_value, max_value)

    def tweak_records_schema_for_load(self,
                                      records_schema: RecordsSchema,
                                      records_format: BaseRecordsFormat) -> RecordsSchema:
        if isinstance(records_format, ParquetRecordsFormat):
            # According to Google, "DATETIME is not supported for
            # uploading from Parquet" -
            # https://github.com/googleapis/google-cloud-python/issues/9996#issuecomment-572273407
            #
            # So we need to make sure we don't create any DATETIME
            # columns if we're loading from a Parquet file.
            #
            # https://cloud.google.com/bigquery/docs/loading-data-cloud-storage-parquet
            return records_schema.cast_field_types({'datetime': 'datetimetz'})
        else:
            return records_schema

    def tweak_records_schema_after_unload(self,
                                          records_schema: RecordsSchema,
                                          records_format: BaseRecordsFormat) -> RecordsSchema:
        if isinstance(records_format, AvroRecordsFormat):
            # https://cloud.google.com/bigquery/docs/loading-data-cloud-storage-avro#logical_types
            #
            # "Note: There is no logical type that directly
            # corresponds to DATETIME, and BigQuery currently doesn't
            # support any direct conversion from an Avro type into a
            # DATETIME field."
            #
            # BigQuery exports this as an Avro string type
            return records_schema.cast_field_types({'datetime': 'string'})
        else:
            return records_schema



# This file is vendored from
# https://raw.githubusercontent.com/jmcarp/sqlalchemy-postgres-copy/01ef522e8e46a6961e227069d465b0cb93e42383/postgres_copy/__init__.py
#
# See https://github.com/jmcarp/sqlalchemy-postgres-copy/pull/14 to
# address the issue on why we had to vendor it instead of creating a
# transitive dpeendency.

import six

from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Mapper, class_mapper
from sqlalchemy.sql.operators import ColumnOperators
from sqlalchemy.dialects import postgresql

__version__ = '0.5.0'


def copy_to(source, dest, engine_or_conn, **flags):
    """Export a query or select to a file. For flags, see the PostgreSQL
    documentation at http://www.postgresql.org/docs/9.5/static/sql-copy.html.

    Examples: ::
        select = MyTable.select()
        with open('/path/to/file.tsv', 'w') as fp:
            copy_to(select, fp, conn)

        query = session.query(MyModel)
        with open('/path/to/file/csv', 'w') as fp:
            copy_to(query, fp, engine, format='csv', null='.')

    :param source: SQLAlchemy query or select
    :param dest: Destination file pointer, in write mode
    :param engine_or_conn: SQLAlchemy engine, connection, or raw_connection
    :param **flags: Options passed through to COPY

    If an existing connection is passed to `engine_or_conn`, it is the caller's
    responsibility to commit and close.
    """
    dialect = postgresql.dialect()
    statement = getattr(source, 'statement', source)
    compiled = statement.compile(dialect=dialect)
    conn, autoclose = raw_connection_from(engine_or_conn)
    cursor = conn.cursor()
    query = cursor.mogrify(compiled.string, compiled.params).decode()
    formatted_flags = '({})'.format(format_flags(flags)) if flags else ''
    copy = 'COPY ({}) TO STDOUT {}'.format(query, formatted_flags)
    cursor.copy_expert(copy, dest)
    if autoclose:
        conn.close()


def copy_from(source, dest, engine_or_conn, columns=(), **flags):
    """Import a table from a file. For flags, see the PostgreSQL documentation
    at http://www.postgresql.org/docs/9.5/static/sql-copy.html.

    Examples: ::
        with open('/path/to/file.tsv') as fp:
            copy_from(fp, MyTable, conn)

        with open('/path/to/file.csv') as fp:
            copy_from(fp, MyModel, engine, format='csv')

    :param source: Source file pointer, in read mode
    :param dest: SQLAlchemy model or table
    :param engine_or_conn: SQLAlchemy engine, connection, or raw_connection
    :param columns: Optional tuple of columns
    :param **flags: Options passed through to COPY

    If an existing connection is passed to `engine_or_conn`, it is the caller's
    responsibility to commit and close.

    The `columns` flag can be set to a tuple of strings to specify the column
    order. Passing `header` alone will not handle out of order columns, it simply tells
    postgres to ignore the first line of `source`.
    """
    tbl = dest.__table__ if is_model(dest) else dest
    conn, autoclose = raw_connection_from(engine_or_conn)
    cursor = conn.cursor()
    relation = '.'.join('"{}"'.format(part) for part in (tbl.schema, tbl.name) if part)
    formatted_columns = '({})'.format(','.join(columns)) if columns else ''
    formatted_flags = '({})'.format(format_flags(flags)) if flags else ''
    copy = 'COPY {} {} FROM STDIN {}'.format(
        relation,
        formatted_columns,
        formatted_flags,
    )
    cursor.copy_expert(copy, source)
    if autoclose:
        conn.commit()
        conn.close()


def raw_connection_from(engine_or_conn):
    """Extract a raw_connection and determine if it should be automatically closed.

    Only connections opened by this package will be closed automatically.
    """
    if hasattr(engine_or_conn, 'cursor'):
        return engine_or_conn, False
    if hasattr(engine_or_conn, 'connection'):
        return engine_or_conn.connection, False
    return engine_or_conn.raw_connection(), True


def format_flags(flags):
    return ', '.join(
        '{} {}'.format(key.upper(), format_flag(value))
        for key, value in flags.items()
    )


def format_flag(value):
    return (
        six.text_type(value).upper()
        if isinstance(value, bool)
        else repr(value)
    )


def relabel_query(query):
    """Relabel query entities according to mappings defined in the SQLAlchemy
    ORM. Useful when table column names differ from corresponding attribute
    names. See http://docs.sqlalchemy.org/en/latest/orm/mapping_columns.html
    for details.

    :param query: SQLAlchemy query
    """
    return query.with_entities(*query_entities(query))


def query_entities(query):
    return sum(
        [desc_entities(desc) for desc in query.column_descriptions],
        []
    )


def desc_entities(desc):
    expr, name = desc['expr'], desc['name']
    if isinstance(expr, Mapper):
        return mapper_entities(expr)
    elif is_model(expr):
        return mapper_entities(expr.__mapper__)
    elif isinstance(expr, ColumnOperators):
        return [expr.label(name)]
    else:
        raise ValueError('Unrecognized query entity {!r}'.format(expr))


def mapper_entities(mapper):
    model = mapper.class_
    return [
        getattr(model, prop.key).label(prop.key)
        for prop in mapper.column_attrs
    ]


def is_model(class_):
    try:
        class_mapper(class_)
        return True
    except SQLAlchemyError:
        return False



from abc import ABCMeta, abstractmethod
from ..records_directory import RecordsDirectory
from ..records_format import BaseRecordsFormat
from ..processing_instructions import ProcessingInstructions
from ..sources import SupportsMoveToRecordsDirectory
from ..results import MoveResult
from typing import Optional, List, TYPE_CHECKING
from ..sources.fileobjs import FileobjsSource
if TYPE_CHECKING:
    from pandas import DataFrame  # noqa
    from ..sources.dataframes import DataframesRecordsSource


class RecordsTarget(metaclass=ABCMeta):
    """These are possible abstract base classes you can implement --
    implement all which can be done efficiently for your
    RecordsTarget.  See move() in mover.py to see how these are used
    or to debug a particular records move scenario).  None are
    required, but a more fully featured implementation is more likely
    to work with the current move() algorithms and allow moving
    records efficiently from a wider range of RecordsSource
    classes."""

    def validate(self) -> None:
        """Raise exception if this target as configured is not valid in some
        way (e.g., directory not empty, DB creds don't work, etc)"""
        pass

    def __str__(self) -> str:
        return type(self).__name__


class NegotiatesRecordsFormat(RecordsTarget, metaclass=ABCMeta):
    @abstractmethod
    def known_supported_records_formats(self) -> List[BaseRecordsFormat]:
        pass

    @abstractmethod
    def can_move_from_format(self,
                             source_records_format: BaseRecordsFormat) -> bool:
        """Returns True if any movement operations can be done using the specified format
        without translation beforehand to a different format"""
        pass


class SupportsRecordsDirectory(NegotiatesRecordsFormat, metaclass=ABCMeta):
    # If a certain format is required when writing to this target, set it here
    records_format: Optional[BaseRecordsFormat]

    def known_supported_records_formats(self) -> List[BaseRecordsFormat]:
        if self.records_format is not None:
            return [self.records_format]
        else:
            return []

    def can_move_from_format(self,
                             source_records_format: BaseRecordsFormat) -> bool:
        """Returns True if any movement operations can be done using the specified format
        without translation beforehand to a different format"""
        return self.records_format is None or self.records_format == source_records_format

    @abstractmethod
    def records_directory(self) -> RecordsDirectory:
        """Represent the current target as a RecordsDirectory object which
        can be filled in.  Note that this is not a context manager, so
        don't implement this method by creating things which need
        cleanup (e.g., creating a temporary directory).  Please see
        to_fileobjs_source() for that situation."""
        pass

    def pre_load_hook(self) -> None:
        """This function will be called before data is loaded into the records
        directory.

        You can expect any exceptions raised by your implementation of
        this hook to be propagated back to the original caller of
        move() - there's no guarantee they will be handled at any
        layer of the move() implementation."""
        pass

    def post_load_hook(self, num_rows_loaded: Optional[int]) -> None:
        """This function will be called after data is loaded into the records
        directory

        You can expect any exceptions raised by your implementation of
        this hook to be propagated back to the original caller of
        move() - there's no guarantee they will be handled at any
        layer of the move() implementation."""
        pass


class SupportsMoveFromRecordsDirectory(NegotiatesRecordsFormat, metaclass=ABCMeta):
    @abstractmethod
    def move_from_records_directory(self,
                                    directory: RecordsDirectory,
                                    processing_instructions: ProcessingInstructions,
                                    override_records_format: Optional[BaseRecordsFormat] = None)\
            -> MoveResult:
        """Given a RecordsDirectory object, load the data inside
        per the ProcessingInstructions and any hint overrides provided."""
        pass

    @abstractmethod
    def can_move_directly_from_scheme(self, scheme: str) -> bool:
        """If true is returned, the load will be done without streaming data
        down to Records Mover byte by byte--which can be expensive
        when data is large and/or network bandwidth is limited.  A
        target that can read from a scheme "directly" in this sense is
        more likely to be efficient in loading."""
        pass


class MightSupportMoveFromFileobjsSource(NegotiatesRecordsFormat, metaclass=ABCMeta):
    @abstractmethod
    def move_from_fileobjs_source(self,
                                  fileobjs_source: FileobjsSource,
                                  processing_instructions: ProcessingInstructions) -> MoveResult:
        pass

    @abstractmethod
    def can_move_from_fileobjs_source(self) -> bool:
        pass


class MightSupportMoveFromTempLocAfterFillingIt(NegotiatesRecordsFormat, metaclass=ABCMeta):
    @abstractmethod
    def can_move_from_temp_loc_after_filling_it(self) -> bool:
        """Returns True if target as currently configured can be handed a
        temporary location and fill it.
        """
        pass

    @abstractmethod
    def temporary_loadable_directory_scheme(self) -> str:
        """Which URL scheme will be used to create the temporary location to fill in."""
        pass

    @abstractmethod
    def move_from_temp_loc_after_filling_it(self,
                                            records_source:
                                            SupportsMoveToRecordsDirectory,
                                            processing_instructions:
                                            ProcessingInstructions) -> MoveResult:
        """Create a temporary location for a RecordsDirectory to live,
        call records_source.move_to_records_directory()
        with it, and then move in the records from the temporary
        RecordsDirectory location.  This can be useful when copying
        records from sources and targets where neither represent
        themselves as a RecordsDirectory (e.g., pandas dataframe
        to/from database), but where the target can load from one if
        it's available.  Not needed if you can implement
        records_directory()."""
        pass


class SupportsMoveFromDataframes(RecordsTarget, metaclass=ABCMeta):
    @abstractmethod
    def move_from_dataframes_source(self,
                                    dfs_source: 'DataframesRecordsSource',
                                    processing_instructions:
                                    ProcessingInstructions) -> MoveResult:
        """
        Given DataFramesRecordsSource, load the data inside per the
        ProcessingInstructions provided.
        """
        pass
